"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import TypeVar, Generic
from .abc import Element, Field, Sendable
from .pile import Pile, pile
from .progression import Progression, progression

T = TypeVar("T")


class Exchange(Element, Generic[T]):
    """
    Item exchange system designed to handle incoming and outgoing flows of items.

    Attributes:
        pile (Pile[T]): The pile of items in the exchange.
        pending_ins (dict[str, Progression]): The pending incoming items to the exchange.
        pending_outs (Progression): The progression of pending outgoing items.
    """

    pile: Pile[T] = Field(
        default_factory=lambda: pile(),
        description="The pile of items in the exchange.",
        title="pending items",
    )

    pending_ins: dict[str, Progression] = Field(
        default_factory=dict,
        description="The pending incoming items to the exchange.",
        title="pending incoming items",
        examples=["{'sender_id': Progression}"],
    )

    pending_outs: Progression = Field(
        default_factory=lambda: progression(),
        description="The pending outgoing items to the exchange.",
        title="pending outgoing items",
    )

    def __contains__(self, item):
        """
        Check if an item is in the pile.

        Args:
            item: The item to check.

        Returns:
            bool: True if the item is in the pile, False otherwise.
        """
        return item in self.pile

    @property
    def unassigned(self) -> Pile[T]:
        """
        if the item is not in the pending_ins or pending_outs, it is unassigned.
        """
        return pile(
            [
                item
                for item in self.pile
                if (
                    all(item not in j for j in self.pending_ins.values())
                    and item not in self.pending_outs
                )
            ]
        )

    @property
    def senders(self) -> list[str]:
        """
        Get the list of senders for the pending incoming items.

        Returns:
            list[str]: The list of sender IDs.
        """
        return list(self.pending_ins.keys())

    def exclude(self, item) -> bool:
        """
        Exclude an item from the exchange.

        Args:
            item: The item to exclude.

        Returns:
            bool: True if the item was successfully excluded, False otherwise.
        """
        return (
            self.pile.exclude(item)
            and all([v.exclude(item) for v in self.pending_ins.values()])
            and self.pending_outs.exclude(item)
        )

    def include(self, item, direction=None) -> bool:
        """
        Include an item in the exchange in a specified direction.

        Args:
            item: The item to include.
            direction (str): The direction to include the item ('in' or 'out').

        Returns:
            bool: True if the item was successfully included, False otherwise.
        """
        if self.pile.include(item):
            item = self.pile[item]
            item = [item] if not isinstance(item, list) else item
            for i in item:
                if not self._include(i, direction=direction):
                    return False
            return True

    def _include(self, item: Sendable, direction) -> bool:
        """
        Helper method to include an item in the exchange in a specified direction.

        Args:
            item (Sendable): The item to include.
            direction (str): The direction to include the item ('in' or 'out').

        Returns:
            bool: True if the item was successfully included, False otherwise.
        """
        if direction == "in":
            if item.sender not in self.pending_ins:
                self.pending_ins[item.sender] = progression()
            return self.pending_ins[item.sender].include(item)

        if direction == "out":
            return self.pending_outs.include(item)

        return True

    def to_dict(self) -> dict:
        """
        Convert the exchange to a dictionary.

        Returns:
            dict: The dictionary representation of the exchange.
        """
        return self.model_dump(by_alias=True)

    def __bool__(self):
        return True

    def __len__(self):
        return len(self.pile)



import unittest
from unittest.mock import MagicMock, patch
import lionagi as li
from lionagi.core.message import System, Instruction, AssistantResponse, ActionResponse
from lionagi.core.collections import Pile, Progression, Exchange
from lionagi.core.action.tool_manager import ToolManager


class TestBranch(unittest.TestCase):

    def setUp(self):
        self.branch = li.Branch()

    def test_initialize_branch(self):
        self.assertIsInstance(self.branch, li.Branch)
        self.assertIsInstance(self.branch.messages, Pile)
        self.assertIsInstance(self.branch.progress, Progression)
        self.assertIsInstance(self.branch.tool_manager, ToolManager)
        self.assertIsInstance(self.branch.mailbox, Exchange)
        self.assertIsInstance(self.branch.imodel, li.iModel)

    def test_add_message_system(self):
        self.branch.add_message(
            system="You are a helpful assistant, let's think step by step"
        )
        self.assertEqual(len(self.branch.messages), 1)
        self.assertEqual(
            self.branch.messages[0].content,
            {"system_info": "You are a helpful assistant, let's think step by step"},
        )

    def test_to_df(self):
        self.branch.add_message(
            system="You are a helpful assistant, let's think step by step"
        )
        df = self.branch.to_df()
        self.assertEqual(df.iloc[0]["message_type"], "System")
        self.assertEqual(df.iloc[0]["role"], "system")

    def test_to_chat_messages(self):
        self.branch.add_message(
            system="You are a helpful assistant, let's think step by step"
        )
        chat_msgs = self.branch.to_chat_messages()
        self.assertEqual(chat_msgs[0]["role"], "system")
        self.assertEqual(
            chat_msgs[0]["content"],
            "You are a helpful assistant, let's think step by step",
        )

    @patch("lionagi.Branch.chat")
    async def test_chat(self, mock_chat):
        mock_chat.return_value = (
            "Rain poured, but their love shone brighter than any storm."
        )
        response = await self.branch.chat("tell me a 10 word story", logprobs=True)
        self.assertEqual(
            response, "Rain poured, but their love shone brighter than any storm."
        )
        mock_chat.assert_called_once_with("tell me a 10 word story", logprobs=True)

    def test_metadata(self):
        self.branch.add_message(
            system="You are a helpful assistant, let's think step by step"
        )
        self.assertIn("last_updated", self.branch.messages[0].metadata)

    # def test_register_tools(self):
    #     tool = MagicMock()
    #     self.branch.register_tools([tool])
    #     self.assertIn(tool, self.branch.tool_manager.registry.values())

    # def test_delete_tools(self):
    #     tool = MagicMock()
    #     tool.schema_ = {"function": {"name": "test_tool"}}
    #     self.branch.register_tools([tool])
    #     self.branch.delete_tools([tool])
    #     self.assertNotIn("test_tool", self.branch.tool_manager.registry)

    def test_send_receive_mail(self):
        self.branch.send = MagicMock()
        self.branch.receive = MagicMock()
        package = MagicMock()
        self.branch.send(recipient="recipient_id", category="message", package=package)
        self.branch.receive(sender="recipient_id")
        self.branch.send.assert_called_once_with(
            recipient="recipient_id", category="message", package=package
        )
        self.branch.receive.assert_called_once_with(sender="recipient_id")

    async def test_chat_with_tool(self):
        async def mock_multiply(number1, number2, number3=1):
            return number1 * number2 * number3

        instruction = """
        solve the following problem
        """
        context = """
        I have 730_000 trees, with average 123 apples per tree, each weigh 0.4 lbs. 
        20 percent are bad and sold for 0.1 dollar per lbs, 30 percent are sold to 
        brewery for 0.3 dollar per apple, what is my revenue?
        """

        self.branch = li.Branch(
            "act like a calculator, invoke tool uses", tools=[mock_multiply]
        )
        response = await self.branch.chat(
            instruction=instruction, context=context, tools=True
        )
        self.assertIsNotNone(response)
        self.assertIn("revenue", response.lower())


if __name__ == "__main__":
    unittest.main()



"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

"""
The base directive module.
"""

import asyncio
import contextlib
import re
from abc import ABC

from typing import Any, Optional

from lionagi.libs import ParseUtil, StringMatch, to_list
from lionagi.libs.ln_nested import nmerge
from lionagi.core.collections.abc import ActionError
from lionagi.core.message import ActionRequest, ActionResponse, Instruction
from lionagi.core.message.util import _parse_action_request
from lionagi.core.report.form import Form
from lionagi.core.unit.util import process_tools
from lionagi.core.validator.validator import Validator


class DirectiveMixin(ABC):
    """
    DirectiveMixin is a class for handling chat operations and
    processing responses.
    """

    def _create_chat_config(
        self,
        system: Optional[str] = None,
        instruction: Optional[str] = None,
        context: Optional[str] = None,
        images: Optional[str] = None,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        requested_fields: Optional[list] = None,
        form: Form = None,
        tools: bool = False,
        branch: Optional[Any] = None,
        **kwargs,
    ) -> Any:
        """
        Create the chat configuration based on the provided parameters.

        Args:
            system: System message.
            instruction: Instruction message.
            context: Context message.
            sender: Sender identifier.
            recipient: Recipient identifier.
            requested_fields: Fields requested in the response.
            form: Form data.
            tools: Flag indicating if tools should be used.
            branch: Branch instance.
            kwargs: Additional keyword arguments.

        Returns:
            dict: The chat configuration.
        """
        branch = branch or self.branch

        if system:
            branch.add_message(system=system)

        if not form:
            if recipient == "branch.ln_id":
                recipient = branch.ln_id

            branch.add_message(
                instruction=instruction,
                context=context,
                sender=sender,
                recipient=recipient,
                requested_fields=requested_fields,
                images=images,
            )
        else:
            instruct_ = Instruction.from_form(form)
            branch.add_message(instruction=instruct_)

        if "tool_parsed" in kwargs:
            kwargs.pop("tool_parsed")
            tool_kwarg = {"tools": tools}
            kwargs = tool_kwarg | kwargs
        elif tools and branch.has_tools:
            kwargs = branch.tool_manager.parse_tool(tools=tools, **kwargs)

        config = {**self.imodel.config, **kwargs}
        if sender is not None:
            config["sender"] = sender

        return config

    async def _call_chatcompletion(
        self, imodel: Optional[Any] = None, branch: Optional[Any] = None, **kwargs
    ) -> Any:
        """
        Calls the chat completion model.

        Args:
            imodel: The model instance.
            branch: The branch instance.
            kwargs: Additional keyword arguments.

        Returns:
            Any: The chat completion result.
        """
        imodel = imodel or self.imodel
        branch = branch or self.branch
        return await imodel.call_chat_completion(branch.to_chat_messages(), **kwargs)

    async def _process_chatcompletion(
        self,
        payload: dict,
        completion: dict,
        sender: str,
        invoke_tool: bool = True,
        branch: Optional[Any] = None,
        action_request: Optional[Any] = None,
        costs=None,
    ) -> Any:
        """
        Processes the chat completion response.
        Currently only support last message for function calling

        Args:
            payload: The payload data.
            completion: The completion data.
            sender: The sender identifier.
            invoke_tool: Flag indicating if tools should be invoked.
            branch: The branch instance.
            action_request: The action request instance.

        Returns:
            Any: The processed result.
        """
        branch = branch or self.branch
        _msg = None

        if "choices" in completion:
            payload.pop("messages", None)
            branch.update_last_instruction_meta(payload)
            _choices = completion.pop("choices", None)

            def process_completion_choice(choice, price=None):
                if isinstance(choice, dict):
                    msg = choice.pop("message", None)
                    _completion = completion.copy()
                    _completion.update(choice)
                    branch.add_message(
                        assistant_response=msg,
                        metadata=_completion,
                        sender=sender,
                    )

                a = branch.messages[-1]._meta_get(
                    ["extra", "usage", "prompt_tokens"], 0
                )
                b = branch.messages[-1]._meta_get(
                    ["extra", "usage", "completion_tokens"], 0
                )
                m = completion.get("model", None)
                if m:
                    ttl = (a * price[0] + b * price[1]) / 1000000
                    branch.messages[-1]._meta_insert(["extra", "usage", "expense"], ttl)
                return msg

            if _choices and not isinstance(_choices, list):
                _choices = [_choices]

            if _choices and isinstance(_choices, list):
                for _choice in _choices:
                    _msg = process_completion_choice(_choice, price=costs)

            branch.imodel.status_tracker.num_tasks_succeeded += 1
        else:
            branch.imodel.status_tracker.num_tasks_failed += 1

        return await self._process_action_request(
            _msg=_msg,
            branch=branch,
            invoke_tool=invoke_tool,
            action_request=action_request,
        )

    async def _process_action_request(
        self,
        _msg: Optional[dict] = None,
        branch: Optional[Any] = None,
        invoke_tool: bool = True,
        action_request: Optional[Any] = None,
    ) -> Any:
        """
        Processes an action request from the assistant response.

        Args:
            _msg: The message data.
            branch: The branch instance.
            invoke_tool: Flag indicating if tools should be invoked.
            action_request: The action request instance.

        Returns:
            Any: The processed result.
        """
        action_request = action_request or _parse_action_request(_msg)
        if action_request is None:
            return _msg if _msg else False

        if action_request:
            for i in action_request:
                if i.function in branch.tool_manager.registry:
                    i.recipient = branch.tool_manager.registry[i.function].ln_id
                else:
                    raise ActionError(f"Tool {i.function} not found in registry")
                branch.add_message(action_request=i, recipient=i.recipient)

        if invoke_tool:
            tasks = []
            for i in action_request:
                tool = branch.tool_manager.registry[i.function]
                tasks.append(asyncio.create_task(tool.invoke(i.arguments)))
            results = await asyncio.gather(*tasks)

            for idx, item in enumerate(results):
                if item is not None:
                    branch.add_message(
                        action_request=action_request[idx],
                        func_outputs=item,
                        sender=action_request[idx].recipient,
                        recipient=action_request[idx].sender,
                    )

        return None

    async def _output(
        self,
        payload: dict,
        completion: dict,
        sender: str,
        invoke_tool: bool,
        requested_fields: dict,
        form: Form = None,
        return_form: bool = True,
        strict: bool = False,
        rulebook: Any = None,
        use_annotation: bool = True,
        template_name: str = None,
        costs=None,
    ) -> Any:
        """
        Outputs the final processed response.

        Args:
            payload: The payload data.
            completion: The completion data.
            sender: The sender identifier.
            invoke_tool: Flag indicating if tools should be invoked.
            requested_fields: Fields requested in the response.
            form: Form data.
            return_form: Flag indicating if form should be returned.
            strict: Flag indicating if strict validation should be applied.
            rulebook: Rulebook instance for validation.
            use_annotation: Flag indicating if annotations should be used.
            template_name: Template name for form.

        Returns:
            Any: The processed response.
        """
        _msg = await self._process_chatcompletion(
            payload=payload,
            completion=completion,
            sender=sender,
            invoke_tool=invoke_tool,
            costs=costs,
        )

        if _msg is None:
            return None

        response_ = self._process_model_response(_msg, requested_fields)

        if form:
            validator = Validator(rulebook=rulebook) if rulebook else self.validator
            form = await validator.validate_response(
                form=form,
                response=response_,
                strict=strict,
                use_annotation=use_annotation,
            )
            if template_name:
                form.template_name = template_name

            return (
                form
                if return_form
                else {
                    i: form.work_fields[i]
                    for i in form.requested_fields
                    if form.work_fields[i] is not None
                }
            )

        return response_

    async def _base_chat(
        self,
        instruction: Any = None,
        *,
        system: Any = None,
        context: Any = None,
        sender: Any = None,
        recipient: Any = None,
        requested_fields: dict = None,
        form: Form = None,
        tools: Any = False,
        images: Optional[str] = None,
        invoke_tool: bool = True,
        return_form: bool = True,
        strict: bool = False,
        rulebook: Any = None,
        imodel: Any = None,
        use_annotation: bool = True,
        branch: Any = None,
        clear_messages: bool = False,
        return_branch: bool = False,
        **kwargs,
    ) -> Any:
        """
        Handles the base chat operation by configuring the chat and
        processing the response.

        Args:
            instruction: Instruction message.
            system: System message.
            context: Context message.
            sender: Sender identifier.
            recipient: Recipient identifier.
            requested_fields: Fields requested in the response.
            form: Form data.
            tools: Flag indicating if tools should be used.
            invoke_tool: Flag indicating if tools should be invoked.
            return_form: Flag indicating if form should be returned.
            strict: Flag indicating if strict validation should be applied.
            rulebook: Rulebook instance for validation.
            imodel: Model instance.
            use_annotation: Flag indicating if annotations should be used.
            branch: Branch instance.
            clear_messages: Flag indicating if messages should be cleared.
            return_branch: Flag indicating if branch should be returned.
            kwargs: Additional keyword arguments.

        Returns:
            Any: The processed response and branch.
        """
        branch = branch or self.branch
        if clear_messages:
            branch.clear()
            branch.set_system(system)

        config = self._create_chat_config(
            system=system,
            instruction=instruction,
            context=context,
            sender=sender,
            recipient=recipient,
            requested_fields=requested_fields,
            form=form,
            tools=tools,
            branch=branch,
            images=images,
            **kwargs,
        )

        payload, completion = await self._call_chatcompletion(
            imodel=imodel, branch=branch, **config
        )

        imodel = imodel or self.imodel
        out_ = await self._output(
            payload=payload,
            completion=completion,
            sender=sender,
            invoke_tool=invoke_tool,
            requested_fields=requested_fields,
            form=form,
            return_form=return_form,
            strict=strict,
            rulebook=rulebook,
            use_annotation=use_annotation,
            costs=imodel.costs,
        )

        return out_, branch if return_branch else out_

    async def _chat(
        self,
        instruction=None,
        context=None,
        system=None,
        sender=None,
        recipient=None,
        branch=None,
        requested_fields=None,
        form: Form = None,
        tools=False,
        invoke_tool=True,
        return_form=True,
        strict=False,
        rulebook=None,
        imodel=None,
        images: Optional[str] = None,
        clear_messages=False,
        use_annotation=True,
        timeout: float = None,
        return_branch=False,
        **kwargs,
    ):
        """
        Handles the chat operation.

        Args:
            instruction: Instruction message.
            context: Context message.
            system: System message.
            sender: Sender identifier.
            recipient: Recipient identifier.
            branch: Branch instance.
            requested_fields: Fields requested in the response.
            form: Form data.
            tools: Flag indicating if tools should be used.
            invoke_tool: Flag indicating if tools should be invoked.
            return_form: Flag indicating if form should be returned.
            strict: Flag indicating if strict validation should be applied.
            rulebook: Rulebook instance for validation.
            imodel: Model instance.
            clear_messages: Flag indicating if messages should be cleared.
            use_annotation: Flag indicating if annotations should be used.
            timeout: Timeout value.
            return_branch: Flag indicating if branch should be returned.
            kwargs: Additional keyword arguments.

        Returns:
            Any: The processed response.
        """
        a = await self._base_chat(
            context=context,
            instruction=instruction,
            system=system,
            sender=sender,
            recipient=recipient,
            requested_fields=requested_fields,
            form=form,
            tools=tools,
            images=images,
            invoke_tool=invoke_tool,
            return_form=return_form,
            strict=strict,
            rulebook=rulebook,
            imodel=imodel,
            use_annotation=use_annotation,
            timeout=timeout,
            branch=branch,
            clear_messages=clear_messages,
            return_branch=return_branch,
            **kwargs,
        )

        if isinstance(a, str):
            return a

        a = list(a)

        if len(a) == 2 and a[0] == a[1]:
            return a[0] if not isinstance(a[0], tuple) else a[0][0]
        if len(a) == 2 and a[0] != a[1]:
            return a[0], a[1]
        if len(a) == 1 and isinstance(a[0], tuple):
            return a[0][0]
        if len(a) == 1 and not isinstance(a[0], tuple):
            return a[0]

    async def _direct(
        self,
        instruction=None,
        context=None,
        form: Form = None,
        branch=None,
        tools=None,
        reason: bool = None,
        predict: bool = None,
        score: bool = None,
        select: bool = None,
        plan: bool = None,
        allow_action: bool = None,
        allow_extension: bool = None,
        confidence: bool = None,
        max_extension: int = None,
        score_num_digits=None,
        score_range=None,
        select_choices=None,
        plan_num_step=None,
        predict_num_sentences=None,
        clear_messages=False,
        return_branch=False,
        images: Optional[str] = None,
        verbose=None,
        **kwargs,
    ):
        """
        Directs the operation based on the provided parameters.

        Args:
            instruction: Instruction message.
            context: Context message.
            form: Form data.
            branch: Branch instance.
            tools: Tools data.
            reason: Flag indicating if reason should be included.
            predict: Flag indicating if prediction should be included.
            score: Flag indicating if score should be included.
            select: Flag indicating if selection should be included.
            plan: Flag indicating if plan should be included.
            allow_action: Flag indicating if action should be allowed.
            allow_extension: Flag indicating if extension should be allowed.
            confidence: Flag indicating if confidence should be included.
            max_extension: Maximum extension value.
            score_num_digits: Number of digits for score.
            score_range: Range for score.
            select_choices: Choices for selection.
            plan_num_step: Number of steps for plan.
            predict_num_sentences: Number of sentences for prediction.
            clear_messages: Flag indicating if messages should be cleared.
            return_branch: Flag indicating if branch should be returned.
            kwargs: Additional keyword arguments.

        Returns:
            Any: The processed response and branch.
        """
        a = await self._base_direct(
            instruction=instruction,
            context=context,
            form=form,
            branch=branch,
            tools=tools,
            reason=reason,
            predict=predict,
            score=score,
            select=select,
            images=images,
            plan=plan,
            allow_action=allow_action,
            allow_extension=allow_extension,
            confidence=confidence,
            max_extension=max_extension,
            score_num_digits=score_num_digits,
            score_range=score_range,
            select_choices=select_choices,
            plan_num_step=plan_num_step,
            predict_num_sentences=predict_num_sentences,
            clear_messages=clear_messages,
            return_branch=return_branch,
            verbose=verbose,
            **kwargs,
        )

        a = list(a)
        if len(a) == 2 and a[0] == a[1]:
            return a[0] if not isinstance(a[0], tuple) else a[0][0]

        return a[0], a[1]

    async def _base_direct(
        self,
        instruction=None,
        *,
        context=None,
        form: Form = None,
        branch=None,
        tools=None,
        reason: bool = None,
        predict: bool = None,
        score: bool = None,
        select: bool = None,
        plan: bool = None,
        allow_action: bool = None,
        allow_extension: bool = None,
        confidence: bool = None,
        max_extension: int = None,
        score_num_digits=None,
        score_range=None,
        select_choices=None,
        plan_num_step=None,
        predict_num_sentences=None,
        clear_messages=False,
        return_branch=False,
        images: Optional[str] = None,
        verbose=None,
        **kwargs,
    ):
        """
        Handles the base direct operation.

        Args:
            instruction: Instruction message.
            context: Context message.
            form: Form data.
            branch: Branch instance.
            tools: Tools data.
            reason: Flag indicating if reason should be included.
            predict: Flag indicating if prediction should be included.
            score: Flag indicating if score should be included.
            select: Flag indicating if selection should be included.
            plan: Flag indicating if plan should be included.
            allow_action: Flag indicating if action should be allowed.
            allow_extension: Flag indicating if extension should be allowed.
            confidence: Flag indicating if confidence should be included.
            max_extension: Maximum extension value.
            score_num_digits: Number of digits for score.
            score_range: Range for score.
            select_choices: Choices for selection.
            plan_num_step: Number of steps for plan.
            predict_num_sentences: Number of sentences for prediction.
            clear_messages: Flag indicating if messages should be cleared.
            return_branch: Flag indicating if branch should be returned.
            kwargs: Additional keyword arguments.

        Returns:
            Any: The processed response and branch.
        """
        # Ensure branch is initialized
        branch = branch or self.branch
        if clear_messages:
            branch.clear()

        # Set a default max_extension if allow_extension is True and max_extension is None
        if allow_extension and not max_extension:
            max_extension = 3  # Set a default limit for recursion

        # Process tools if provided
        if tools:
            process_tools(tools, branch)

        if allow_action and not tools:
            tools = True

        tool_schema=None
        if tools:
            tool_schema = branch.tool_manager.get_tool_schema(tools)

        if not form:
            form = self.default_template(
                instruction=instruction,
                context=context,
                reason=reason,
                predict=predict,
                score=score,
                select=select,
                plan=plan,
                tool_schema=tool_schema,
                allow_action=allow_action,
                allow_extension=allow_extension,
                max_extension=max_extension,
                confidence=confidence,
                score_num_digits=score_num_digits,
                score_range=score_range,
                select_choices=select_choices,
                plan_num_step=plan_num_step,
                predict_num_sentences=predict_num_sentences,
            )

        elif form and "tool_schema" not in form._all_fields:
            form.append_to_input("tool_schema")
            form.tool_schema = tool_schema

        else:
            form.tool_schema = tool_schema

        verbose = (
            verbose
            if verbose is not None and isinstance(verbose, bool)
            else self.verbose
        )
        if verbose:
            print("Chatting with model...")

        # Call the base chat method
        form = await self._chat(
            form=form,
            branch=branch,
            images=images,
            **kwargs,
        )

        # Handle actions if allowed and required
        if allow_action and getattr(form, "action_required", None):
            actions = getattr(form, "actions", None)
            if actions:
                if verbose:
                    print(
                        "Found action requests in model response. Processing actions..."
                    )
                form = await self._act(form, branch, actions=actions)
                if verbose:
                    print("Actions processed!")

        last_form = form

        ctr = 1

        # Handle extensions if allowed and required
        extension_forms = []
        max_extension = max_extension if isinstance(max_extension, int) else 3
        while (
            allow_extension
            and max_extension > 0
            and getattr(last_form, "extension_required", None)
        ):
            if getattr(last_form, "is_extension", None):
                break
            if verbose:
                print(f"\nFound extension requests in model response.")
                print(
                    f"------------------- Processing extension No.{ctr} -------------------"
                )

            max_extension -= 1

            # new form cannot be extended, otherwise it will be an infinite loop
            new_form = await self._extend(
                tools=tools,
                reason=reason,
                predict=predict,
                score=score,
                select=select,
                plan=getattr(last_form, "plan", None),
                allow_action=allow_action,
                confidence=confidence,
                score_num_digits=score_num_digits,
                score_range=score_range,
                select_choices=select_choices,
                predict_num_sentences=predict_num_sentences,
                **kwargs,
            )

            if verbose:
                print(f"------------------- Extension completed -------------------\n")

            extension_forms.extend(new_form)
            last_form = new_form[-1] if isinstance(new_form, list) else new_form
            ctr += len(form)

        if extension_forms:
            if not getattr(form, "extension_forms", None):
                form._add_field("extension_forms", list, None, [])
            form.extension_forms.extend(extension_forms)
            action_responses = [
                i.action_response
                for i in extension_forms
                if getattr(i, "action_response", None) is not None
            ]
            if not hasattr(form, "action_response"):
                form.add_field("action_response", {})

            for action_response in action_responses:
                nmerge([form.action_response, action_response])

        if "PLEASE_ACTION" in form.answer:
            if verbose:
                print("Analyzing action responses and generating answer...")

            answer = await self._chat(
                "please provide final answer basing on the above"
                " information, provide answer value as a string only"
                " do not return as json, do not include other information",
            )

            if isinstance(answer, dict):
                a = answer.get("answer", None)
                if a is not None:
                    answer = a

            answer = str(answer).strip()
            if answer.startswith("{") and answer.endswith("}"):
                answer = answer[1:-1]
                answer = answer.strip()
            if '"answer":' in answer:
                answer.replace('"answer":', "")
                answer = answer.strip()
            elif "'answer':" in answer:
                answer.replace("'answer':", "")
                answer = answer.strip()

            form.answer = answer

        return form, branch if return_branch else form

    async def _extend(
        self,
        tools,
        reason,
        predict,
        score,
        select,
        plan,
        # image,
        allow_action,
        confidence,
        score_num_digits,
        score_range,
        select_choices,
        predict_num_sentences,
        **kwargs,
    ):
        """
        Handles the extension of the form based on the provided parameters.

        Args:
            form: Form data.
            tools: Tools data.
            reason: Flag indicating if reason should be included.
            predict: Flag indicating if prediction should be included.
            score: Flag indicating if score should be included.
            select: Flag indicating if selection should be included.
            plan: Flag indicating if plan should be included.
            allow_action: Flag indicating if action should be allowed.
            confidence: Flag indicating if confidence should be included.
            score_num_digits: Number of digits for score.
            score_range: Range for score.
            select_choices: Choices for selection.
            predict_num_sentences: Number of sentences for prediction.
            allow_extension: Flag indicating if extension should be allowed.
            max_extension: Maximum extension value.
            kwargs: Additional keyword arguments.

        Returns:
            list: The extended forms.
        """
        extension_forms = []

        # Ensure the next step in the plan is handled
        directive_kwargs = {
            "tools": tools,
            "reason": reason,
            "predict": predict,
            "score": score,
            "select": select,
            "allow_action": allow_action,
            "confidence": confidence,
            "score_num_digits": score_num_digits,
            "score_range": score_range,
            "select_choices": select_choices,
            "predict_num_sentences": predict_num_sentences,
            **kwargs,
        }

        if plan:
            keys = [f"step_{i+1}" for i in range(len(plan))]
            plan = StringMatch.force_validate_dict(plan, keys)

            # If plan is provided, process each step
            for i in keys:
                directive_kwargs["instruction"] = plan[i]
                last_form = await self._direct(**directive_kwargs)
                last_form.is_extension = True
                extension_forms.append(last_form)
                directive_kwargs["max_extension"] -= 1
                if not getattr(last_form, "extension_required", None):
                    break

        else:
            # Handle single step extension
            last_form = await self._direct(**directive_kwargs)
            last_form.is_extension = True
            extension_forms.append(last_form)

        return extension_forms

    async def _act(self, form, branch, actions=None):
        """
        Processes actions based on the provided form and actions.

        Args:
            form: Form data.
            branch: Branch instance.
            actions: Actions data.

        Returns:
            dict: The updated form.
        """
        if getattr(form, "action_performed", None) is True:
            return form

        keys = [f"action_{i+1}" for i in range(len(actions))]
        actions = StringMatch.force_validate_dict(actions, keys)

        try:
            requests = []
            for k in keys:
                _func = actions[k]["function"]
                _func = _func.replace("functions.", "")
                msg = ActionRequest(
                    function=_func,
                    arguments=actions[k]["arguments"],
                    sender=branch.ln_id,
                    recipient=branch.tool_manager.registry[_func].ln_id,
                )
                requests.append(msg)
                branch.add_message(action_request=msg)

            if requests:
                out = await self._process_action_request(
                    branch=branch, invoke_tool=True, action_request=requests
                )

                if out is False:
                    raise ValueError("No requests found.")

                len_actions = len(actions)
                action_responses = [
                    i
                    for i in branch.messages[-len_actions:]
                    if isinstance(i, ActionResponse)
                ]

                _action_responses = {}
                for idx, item in enumerate(action_responses):
                    _action_responses[f"action_{idx+1}"] = item._to_dict()

                form.append_to_request("action_response")
                if (a := getattr(form, "action_response", None)) is None:
                    form.add_field("action_response", {})

                len1 = len(form.action_response)
                for k, v in _action_responses.items():
                    while k in form.action_response:
                        k = f"{k}_1"
                    form.action_response[k] = v

                if len(form.action_response) > len1:
                    form.append_to_request("action_performed")
                    form.action_performed = True
                return form

        except Exception as e:
            raise ValueError(f"Error processing action request: {e}")

    async def _select(
        self,
        form=None,
        choices=None,
        reason=False,
        confidence_score=None,
        instruction=None,
        template=None,
        context=None,
        branch=None,
        **kwargs,
    ):
        """
        Selects a response based on the provided parameters.

        Args:
            form (Any, optional): Form to create instruction from.
            choices (Any, optional): Choices for the selection.
            reason (bool, optional): Whether to include a reason for the selection.
            confidence_score (Any, optional): Confidence score for the selection.
            instruction (Any, optional): Instruction for the selection.
            template (Any, optional): Template for the selection.
            context (Any, optional): Context to perform the selection on.
            branch (Any, optional): Branch to use for the selection.
            **kwargs: Additional arguments for the selection.

        Returns:
            Any: The selection response.
        """
        branch = branch or self.branch

        if not form:
            form = template(
                choices=choices,
                reason=reason,
                confidence_score=confidence_score,
                instruction=instruction,
                context=context,
            )

        return await self._chat(form=form, return_form=True, branch=branch, **kwargs)

    async def _predict(
        self,
        form=None,
        num_sentences=None,
        reason=False,
        confidence_score=None,
        instruction=None,
        context=None,
        branch=None,
        template=None,
        **kwargs,
    ):
        """
        Predicts a response based on the provided parameters.

        Args:
            form: Form data.
            num_sentences: Number of sentences for the prediction.
            reason: Flag indicating if reason should be included.
            confidence_score: Confidence score for the prediction.
            instruction: Instruction for the prediction.
            context: Context to perform the prediction on.
            branch: Branch instance.
            template: Template for the prediction.
            kwargs: Additional keyword arguments.

        Returns:
            Any: The prediction response.
        """
        branch = branch or self.branch

        if not form:
            form = template(
                instruction=instruction,
                context=context,
                num_sentences=num_sentences,
                confidence_score=confidence_score,
                reason=reason,
            )

        return await self._chat(form=form, return_form=True, branch=branch, **kwargs)

    async def _score(
        self,
        form=None,
        score_range=None,
        include_endpoints=None,
        num_digit=None,
        reason=False,
        confidence_score=None,
        instruction=None,
        context=None,
        branch=None,
        template=None,
        **kwargs,
    ):
        """
        Scores a response based on the provided parameters.

        Args:
            form: Form data.
            score_range: Range for score.
            include_endpoints: Flag indicating if endpoints should be included.
            num_digit: Number of digits for score.
            reason: Flag indicating if reason should be included.
            confidence_score: Confidence score for the score.
            instruction: Instruction for the score.
            context: Context to perform the score on.
            branch: Branch instance.
            template: Template for the score.
            kwargs: Additional keyword arguments.

        Returns:
            Any: The score response.
        """
        branch = branch or self.branch
        if not form:
            form = template(
                score_range=score_range,
                include_endpoints=include_endpoints,
                num_digit=num_digit,
                reason=reason,
                confidence_score=confidence_score,
                instruction=instruction,
                context=context,
            )

        return await self._chat(form=form, return_form=True, branch=branch, **kwargs)

    async def _plan(
        self,
        form=None,
        num_step=None,
        reason=False,
        confidence_score=None,
        instruction=None,
        context=None,
        branch=None,
        template=None,
        **kwargs,
    ):
        """
        Plans a response based on the provided parameters.

        Args:
            form: Form data.
            num_step: Number of steps for the plan.
            reason: Flag indicating if reason should be included.
            confidence_score: Confidence score for the plan.
            instruction: Instruction for the plan.
            context: Context to perform the plan on.
            branch: Branch instance.
            template: Template for the plan.
            kwargs: Additional keyword arguments.

        Returns:
            Any: The plan response.
        """
        branch = branch or self.branch
        template = template or self.default_template

        if not form:
            form = template(
                instruction=instruction,
                context=context,
                num_step=num_step,
                reason=reason,
                confidence_score=confidence_score,
            )

        return await self._chat(form=form, **kwargs)

    @staticmethod
    def _process_model_response(content_, requested_fields):
        """
        Processes the model response content.

        Args:
            content_: The content data.
            requested_fields: Fields requested in the response.

        Returns:
            Any: The processed response.
        """
        out_ = content_.get("content", "")
        if out_ == "":
            out_ = content_

        if requested_fields:
            with contextlib.suppress(Exception):
                return StringMatch.force_validate_dict(out_, requested_fields)

        if isinstance(out_, str):
            with contextlib.suppress(Exception):
                return ParseUtil.fuzzy_parse_json(out_)

            with contextlib.suppress(Exception):
                return ParseUtil.extract_json_block(out_)

            with contextlib.suppress(Exception):
                match = re.search(r"```json\n({.*?})\n```", out_, re.DOTALL)
                if match:
                    return ParseUtil.fuzzy_parse_json(match.group(1))

        return out_



import asyncio
import unittest
from unittest.mock import patch
from lionagi.libs.ln_queue import AsyncQueue


class TestAsyncQueue(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)
        self.queue = AsyncQueue(max_concurrent_tasks=3)

    def tearDown(self):
        self.loop.close()

    def test_enqueue_dequeue(self):
        async def test():
            await self.queue.enqueue("task1")
            result = await self.queue.dequeue()
            self.assertEqual(result, "task1")

        self.loop.run_until_complete(test())

    def test_stop(self):
        async def test():
            await self.queue.stop()
            self.assertTrue(self.queue.stopped())

        self.loop.run_until_complete(test())

    @patch("lionagi.libs.ln_func_call.rcall", autospec=True)
    def test_process_requests(self, mock_rcall):
        future = asyncio.Future()
        future.set_result("Processed successfully")
        mock_rcall.return_value = future

        async def processor(task):
            return f"Processed {task}"

        async def add_tasks():
            for i in range(5):
                await self.queue.enqueue(f"task{i}")
            # This ensures that the queue stops after all tasks are added
            await self.queue.stop()

        async def process():
            await self.queue.process_requests(processor, retry_kwargs={"retries": 2})

        self.loop.create_task(add_tasks())
        self.loop.run_until_complete(process())
        # mock_rcall.assert_called()

    def test_concurrency_limit(self):
        async def add_tasks():
            for i in range(10):
                await self.queue.enqueue(f"task{i}")

        async def process():
            await asyncio.sleep(0.1)
            await self.queue.stop()

        self.loop.run_until_complete(add_tasks())
        self.loop.run_until_complete(process())


if __name__ == "__main__":
    unittest.main()



"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from lionagi.libs.ln_convert import to_str
from lionagi.core.collections.abc import Field
from lionagi.core.unit import UnitForm


class ScoreTemplate(UnitForm):
    """
    A template for performing a scoring task based on given instructions and context.

    Inherits from `UnitForm` and adds fields and methods specific to scoring tasks.

    Attributes:
        confidence_score (float | None): A numeric score between 0 to 1 formatted to 2 decimal places, indicating confidence in the score.
        reason (str | None): A brief reason for the given output.
        template_name (str): The name of the template.
        score (float | None): A score for the given context and task.
        assignment (str): A string representing the task assignment for the score.
    """

    confidence_score: float | None = Field(
        None,
        description="a numeric score between 0 to 1 formatted in num:0.2f, 1 being very confident and 0 being not confident at all",
        validation_kwargs={
            "upper_bound": 1,
            "lower_bound": 0,
            "num_type": float,
            "precision": 2,
        },
    )

    reason: str | None = Field(
        None,
        description="brief reason for the given output, format: This is my best response because ...",
    )

    template_name: str = "score_template"

    score: float | None = Field(
        None,
        description="a score for the given context and task, numeric",
    )

    assignment: str = "task -> score"

    @property
    def answer(self):
        """
        Gets the score attribute.

        Returns:
            float | None: The score for the given context and task.
        """
        return getattr(self, "score", None)

    def __init__(
        self,
        *,
        instruction=None,
        context=None,
        score_range=(0, 10),
        include_endpoints=True,
        num_digit=0,
        confidence_score=False,
        reason=False,
        **kwargs,
    ):
        """
        Initializes a new instance of the ScoreTemplate class.

        Args:
            instruction (str, optional): Additional instructions for the scoring task.
            context (str, optional): Additional context for the scoring task.
            score_range (tuple, optional): The range of the score. Defaults to (0, 10).
            include_endpoints (bool, optional): Whether to include the endpoints in the score range. Defaults to True.
            num_digit (int, optional): The number of digits allowed after the decimal point. Defaults to 0.
            confidence_score (bool, optional): Whether to include a confidence score. Defaults to False.
            reason (bool, optional): Whether to include a reasoning field. Defaults to False.
            **kwargs: Additional keyword arguments.
        """
        super().__init__(**kwargs)

        return_precision = ""

        if num_digit == 0:
            return_precision = "integer"
        else:
            return_precision = f"num:{to_str(num_digit)}f"

        self.task = f"""
perform scoring task according to the following constraints:
1. additional objective: {to_str(instruction or "N/A")}.
2. score range: {to_str(score_range)}.
3. include_endpoints: {"yes" if include_endpoints else "no"}.
4. precision, (max number of digits allowed after "."): {return_precision}.
5. additional information: {to_str(context or "N/A")}.
"""
        if reason:
            self.append_to_request("reason")

        if confidence_score:
            self.append_to_request("confidence_score")

        self.validation_kwargs["score"] = {
            "upper_bound": score_range[1],
            "lower_bound": score_range[0],
            "num_type": int if num_digit == 0 else float,
            "precision": num_digit if num_digit != 0 else None,
        }



"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import asyncio
from lionagi.core.work.work import WorkStatus


class WorkQueue:
    """
    A class representing a queue for managing work.

    Attributes:
        capacity (int): The maximum number of tasks the queue can handle.
        queue (asyncio.Queue): The queue holding the tasks.
        _stop_event (asyncio.Event): Event to signal stopping the execution of the queue.
        available_capacity (int): The remaining number of tasks the queue can handle.
        execution_mode (bool): If `execute` is running.
        refresh_time (int): The time interval between task processing.
    """

    def __init__(self, capacity=5, refresh_time=1):
        """
        Initializes a new instance of WorkQueue.

        Args:
            capacity (int): The maximum number of tasks the queue can handle.
            refresh_time (int): The time interval between task processing.

        Raises:
            ValueError: If capacity is less than 0 or refresh_time is negative.
        """
        if capacity < 0:
            raise ValueError("initial capacity must be >= 0")
        if refresh_time < 0:
            raise ValueError("refresh time for execution can not be negative")
        self.capacity = capacity
        self.queue = asyncio.Queue()
        self._stop_event = asyncio.Event()
        self.available_capacity = capacity
        self.execution_mode = False
        self.refresh_time = refresh_time

    async def enqueue(self, work) -> None:
        """Enqueue a work item."""
        await self.queue.put(work)

    async def dequeue(self):
        """Dequeue a work item."""
        return await self.queue.get()

    async def join(self) -> None:
        """Block until all items in the queue have been processed."""
        await self.queue.join()

    async def stop(self) -> None:
        """Signal the queue to stop processing."""
        self._stop_event.set()

    @property
    def stopped(self) -> bool:
        """Return whether the queue has been stopped."""
        return self._stop_event.is_set()

    async def process(self) -> None:
        """Process the work items in the queue."""
        tasks = set()
        while self.available_capacity > 0 and self.queue.qsize() > 0:
            next = await self.dequeue()
            next.status = WorkStatus.IN_PROGRESS
            task = asyncio.create_task(next.perform())
            tasks.add(task)
            self.available_capacity -= 1

        if tasks:
            await asyncio.wait(tasks)
            self.available_capacity = self.capacity

    async def execute(self):
        """
            Continuously executes the process method at a specified refresh interval.

            Args:
                refresh_time (int, optional): The time in seconds to wait between
                    successive calls to `process`. Defaults to 1.
        """
        self.execution_mode = True
        self._stop_event.clear()

        while not self.stopped:
            await self.process()
            await asyncio.sleep(self.refresh_time)
        self.execution_mode = False



"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Callable, Union, List, Dict, Any
from pydantic import Field, field_serializer
from lionagi.libs.ln_func_call import call_handler
from lionagi.core.collections.abc import Actionable
from lionagi.core.generic.node import Node
from .function_calling import FunctionCalling


class Tool(Node, Actionable):
    """
    Class representing a Tool with capabilities for pre-processing, post-processing,
    and parsing function results.

    Attributes:
        function (Callable): The callable function or capability of the tool.
        schema_ (Union[Dict, None]): Schema in OpenAI format.
        pre_processor (Union[Callable, None]): Function to preprocess input arguments.
        pre_processor_kwargs (Union[Dict, None]): Keyword arguments for the pre-processor.
        post_processor (Union[Callable, None]): Function to post-process the result.
        post_processor_kwargs (Union[Dict, None]): Keyword arguments for the post-processor.
        parser (Union[Callable, None]): Function to parse the result to a JSON serializable format.
    """

    function: Callable = Field(
        ...,
        description="The callable function or capability of the tool.",
    )

    schema_: Union[Dict, None] = Field(
        None,
        description="Schema in OpenAI format.",
    )

    pre_processor: Union[Callable, None] = None
    pre_processor_kwargs: Union[Dict, None] = None

    post_processor: Union[Callable, None] = None
    post_processor_kwargs: Union[Dict, None] = None

    parser: Union[Callable, None] = None  # Parse result to JSON serializable format

    @field_serializer("function", check_fields=False)
    def serialize_func(self, func: Callable) -> str:
        """
        Serialize the function for storage or transmission.

        Args:
            func (Callable): The function to serialize.

        Returns:
            str: The name of the function.
        """
        return func.__name__

    @property
    def name(self) -> str:
        """
        Get the name of the function from the schema.

        Returns:
            str: The name of the function.
        """
        return self.schema_["function"]["name"]

    async def invoke(self, kwargs: Dict = {}) -> Any:
        """
        Invoke the tool's function with optional pre-processing and post-processing.

        Args:
            kwargs (Dict): The arguments to pass to the function.

        Returns:
            Any: The result of the function call.
        """
        if self.pre_processor:
            pre_process_kwargs = self.pre_processor_kwargs or {}
            kwargs = await call_handler(
                self.pre_processor(kwargs, **pre_process_kwargs)
            )
            if not isinstance(kwargs, dict):
                raise ValueError("Pre-processor must return a dictionary.")

        func_call_ = FunctionCalling(function=self.function, arguments=kwargs)
        try:
            result = await func_call_.invoke()
        except Exception:
            return None

        if self.post_processor:
            post_process_kwargs = self.post_processor_kwargs or {}
            result = await call_handler(
                self.post_processor(result, **post_process_kwargs)
            )

        return result if not self.parser else self.parser(result)


TOOL_TYPE = Union[bool, Tool, str, List[Union[Tool, str, Dict]], Dict]



"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from enum import Enum
from lionagi.libs.ln_convert import to_str, to_dict
from lionagi.core.collections.abc import Field
from .template.base import BaseUnitForm


class UnitForm(BaseUnitForm):
    """
    Form for managing unit directives and outputs.

    Attributes:
        actions (dict|None): Actions to take based on the context and instruction.
        action_required (bool|None): Set to True if actions are provided, else False.
        answer (str|None): Answer to the questions asked.
        extension_required (bool|None): Set to True if more steps are needed.
        prediction (str|None): Likely prediction based on context and instruction.
        plan (dict|str|None): Step-by-step plan.
        next_steps (dict|str|None): Ideas on next actions to take.
        score (float|None): Numeric performance score.
        reflection (str|None): Self-reflection on reasoning and performance.
        selection (Enum|str|list|None): A single item from the choices.
        tool_schema (list|dict|None): The list of tools available for use.
        assignment (str): Default assignment task description.
    """

    actions: dict | None = Field(
        None,
        description=(
            "Actions to take based on the context and instruction. "
            "Format: {action_n: {function: ..., arguments: {param1: ..., "
            "param2: ...}}}. Leave blank if no actions are needed. "
            "Must use provided functions and parameters, DO NOT MAKE UP NAMES!!! "
            "Flag `action_required` as True if filled. When providing parameters, "
            "you must follow the provided type and format. If the parameter is a "
            "number, you should provide a number like 1, 23, or 1.1 if float is "
            "allowed."
        ),
        examples=["{action_1: {function: 'add', arguments: {num1: 1, num2: 2}}}"],
    )

    action_required: bool | None = Field(
        None,
        description="Set to True if you provide actions, or provide actions if True.",
        examples=[True, False],
    )

    answer: str | None = Field(
        None,
        description=(
            "Adhere to the prompt and all user instructions. Provide the answer "
            "for the task. if actions are required at this step, set "
            "`action_required` to True and write only `PLEASE_ACTION` to the answer_field."
            "Additionally, if extensions are allowed and needed at this step to provide a"
            " high-quality, accurate answer, set extension_required to True and"
            "you will have another chance to provide the answer after the actions are done.",
        ),
    )

    extension_required: bool | None = Field(
        None,
        description=(
            "Set to True if more steps are needed to provide an accurate answer. "
            "If True, additional rounds are allowed."
        ),
        examples=[True, False],
    )

    prediction: None | str = Field(
        None,
        description="Provide the likely prediction based on context and instruction.",
    )

    plan: dict | str | None = Field(
        None,
        description=(
            "Provide a step-by-step plan. Format: {step_n: {plan: ..., reason: ...}}. "
            "Achieve the final answer at the last step. Set `extend_required` to True "
            "if plan requires more steps."
        ),
        examples=["{step_1: {plan: '...', reason: '...'}}"],
    )

    next_steps: dict | str | None = Field(
        None,
        description=(
            "Brainstorm ideas on next actions to take. Format: {next_step_n: {plan: "
            "..., reason: ...}}. Next_step is about anticipating future actions, "
            "but it does not have to be in a sequential format. Set `extend_required` "
            "to True if more steps are needed."
        ),
        examples=["{next_step_1: {plan: '...', reason: '...'}}"],
    )

    score: float | None = Field(
        None,
        description=(
            "A numeric score. Higher is better. If not otherwise instructed, "
            "fill this field with your own performance rating. Try hard and be "
            "self-critical."
        ),
        examples=[0.2, 5, 2.7],
    )

    reflection: str | None = Field(
        None,
        description=(
            "Reason your own reasoning. Create specific items on how/where/what "
            "you could improve to better achieve the task, or can the problem be "
            "solved in a different and better way. If you can think of a better "
            "solution, please provide it and fill the necessary fields in "
            "`action_required`, `extension_required`, `next_steps` if appropriate."
        ),
    )

    selection: Enum | str | list | None = Field(
        None, description="a single item from the choices."
    )

    tool_schema: list | dict | None = Field(
        None, description="The list of tools available for using."
    )

    assignment: str = Field("task -> answer")

    is_extension: bool = Field(False)

    def __init__(
        self,
        instruction=None,
        context=None,
        reason: bool = True,
        predict: bool = False,
        score=True,
        select=None,
        plan=None,
        brainstorm=None,
        reflect=None,
        tool_schema=None,
        allow_action: bool = False,
        allow_extension: bool = False,
        max_extension: int = None,
        confidence=None,
        score_num_digits=None,
        score_range=None,
        select_choices=None,
        plan_num_step=None,
        predict_num_sentences=None,
        **kwargs,
    ):
        """
        Initialize the UnitForm with various parameters and settings.

        Args:
            instruction (str|None): Additional instruction.
            context (str|None): Additional context.
            reason (bool): Flag to include reasoning.
            predict (bool): Flag to include prediction.
            score (bool): Flag to include score.
            select (Enum|str|list|None): Selection choices.
            plan (dict|str|None): Step-by-step plan.
            brainstorm (bool): Flag to include brainstorming next steps.
            reflect (bool): Flag to include self-reflection.
            tool_schema (list|dict|None): Schema of available tools.
            allow_action (bool): Allow actions to be added.
            allow_extension (bool): Allow extension for more steps.
            max_extension (int|None): Maximum number of extensions allowed.
            confidence (bool|None): Flag to include confidence score.
            score_num_digits (int|None): Number of decimal places for the score.
            score_range (list|None): Range for the score.
            select_choices (list|None): Choices for selection.
            plan_num_step (int|None): Number of steps in the plan.
            predict_num_sentences (int|None): Number of sentences for prediction.
            **kwargs: Additional keyword arguments.
        """
        super().__init__(**kwargs)

        self.task = (
            f"Follow the prompt and provide the necessary output.\n"
            f"- Additional instruction: {to_str(instruction or 'N/A')}\n"
            f"- Additional context: {to_str(context or 'N/A')}\n"
        )

        if reason:
            self.append_to_request("reason")

        if allow_action:
            self.append_to_request("actions, action_required, reason")
            self.task += "- Reason and prepare actions with GIVEN TOOLS ONLY.\n"

        if allow_extension:
            self.append_to_request("extension_required")
            self.task += (
                f"- Allow auto-extension up to another {max_extension} rounds.\n"
            )

        if tool_schema:
            self.append_to_input("tool_schema")
            self.tool_schema = tool_schema

        if brainstorm:
            self.append_to_request("next_steps, extension_required")
            self.task += "- Explore ideas on next actions to take.\n"

        if plan:
            plan_num_step = plan_num_step or 3
            max_extension = max_extension or plan_num_step
            allow_extension = True
            self.append_to_request("plan, extension_required")
            self.task += (
                f"- Generate a {plan_num_step}-step plan based on the context.\n"
            )

        if predict:
            self.append_to_request("prediction")
            self.task += (
                f"- Predict the next {predict_num_sentences or 1} sentence(s).\n"
            )

        if select:
            self.append_to_request("selection")
            self.task += (
                f"- Select 1 item from the provided choices: {select_choices}.\n"
            )

        if confidence:
            self.append_to_request("confidence_score")

        if score:
            self.append_to_request("score")

            score_range = score_range or [0, 10]
            score_num_digits = score_num_digits or 0

            self.validation_kwargs["score"] = {
                "upper_bound": score_range[1],
                "lower_bound": score_range[0],
                "num_type": int if score_num_digits == 0 else float,
                "precision": score_num_digits if score_num_digits != 0 else None,
            }

            self.task += (
                f"- Give a numeric score in [{score_range[0]}, {score_range[1]}] "
                f"and precision of {score_num_digits or 0} decimal places.\n"
            )

        if reflect:
            self.append_to_request("reflection")

    def display(self):
        """
        Display the current form fields and values in a user-friendly format.
        """
        fields = self.work_fields.copy()

        if "task" in fields and len(str(fields["task"])) > 2000:
            fields["task"] = fields["task"][:2000] + "..."

        if "tool_schema" in fields:
            tools = to_dict(fields["tool_schema"])["tools"]
            fields["available_tools"] = [i["function"]["name"] for i in tools]
            fields.pop("tool_schema")

        if "actions" in fields:
            a = ""
            idx = 0
            for _, v in fields["actions"].items():
                a += (
                    f"\n \n{idx+1}. **{v['function']}**"
                    f"({', '.join([f'{k}: {v}' for k, v in v['arguments'].items()])}), "
                )
                idx += 1
            fields["actions"] = a[:-2]

        if "action_response" in fields:
            a = ""
            idx = 0
            for _, v in fields["action_response"].items():
                a += (
                    f"\n \n{idx+1}. **{v['function']}**"
                    f"({', '.join([f'{k}: {v}' for k, v in v['arguments'].items()])})"
                )
                if len(str(v["output"])) > 30:
                    a += f" \n \n {v['output']}, "
                else:
                    a += f" = {v['output']}, "
                idx += 1
            fields["action_response"] = a[:-2]

        super().display(fields=fields)



import unittest
from lionagi import Form


class TestForm(unittest.TestCase):

    def setUp(self):
        self.form = Form(assignment="input1, input2 -> output")

    def test_initial_state(self):
        self.assertEqual(self.form.input_fields, ["input1", "input2"])
        self.assertEqual(self.form.requested_fields, ["output"])
        self.assertFalse(self.form.workable)
        self.assertFalse(self.form.filled)
        self.assertEqual(
            self.form.work_fields, {"input1": None, "input2": None, "output": None}
        )

    def test_fill_input_fields(self):
        self.form.fill(input1=1, input2=2)
        self.assertTrue(self.form.workable)
        self.assertFalse(self.form.filled)
        self.assertEqual(
            self.form.work_fields, {"input1": 1, "input2": 2, "output": None}
        )

    def test_fill_all_fields(self):
        self.form.fill(input1=1, input2=2)
        self.form.fill(output=3)
        self.assertTrue(self.form.filled)
        self.assertFalse(self.form.workable)
        self.assertEqual(self.form.work_fields, {"input1": 1, "input2": 2, "output": 3})

    def test_fill_once(self):
        self.form.fill(input1=1, input2=2)
        self.form.fill(output=3)
        with self.assertRaises(ValueError) as context:
            self.form.fill(input1=2, input2=3)
        self.assertTrue(
            "Form is filled, cannot be worked on again" in str(context.exception)
        )
        self.assertEqual(self.form.work_fields, {"input1": 1, "input2": 2, "output": 3})


if __name__ == "__main__":
    unittest.main()



import unittest
from typing import Dict, List, Any
from lionagi.core.report.form import Form
from lionagi.core.report.report import Report
from lionagi.core.collections.abc import FieldError
from lionagi.core.rule.base import Rule
from lionagi.core.rule.rulebook import RuleBook
from lionagi.core.validator.validator import (
    Validator,
    _DEFAULT_RULEORDER,
    _DEFAULT_RULES,
)


class MockRule(Rule):
    def applies(self, *args, **kwargs):
        return True

    async def invoke(self, field, value, form, *args, **kwargs):
        if isinstance(value, int) and value < 0:
            raise ValueError(f"Field {field} cannot be negative.")
        return value

    async def validate(self, value: Any) -> Any:
        return value


class ValidatorTestCase(unittest.TestCase):
    def setUp(self):
        self.validator = Validator(
            rulebook=RuleBook(rules=_DEFAULT_RULES, ruleorder=_DEFAULT_RULEORDER)
        )
        self.form = Form(
            assignment="input1, input2 -> output",
            input_fields=["input1", "input2"],
            requested_fields=["output"],
        )
        self.form2 = Form(
            assignment="total_amount, bike_price -> repair_price",
            input_fields=["total_amount", "bike_price"],
            requested_fields=["repair_price"],
        )

    def test_initiate_rules(self):
        self.assertTrue(isinstance(self.validator.active_rules, dict))

    async def test_validate_field(self):
        self.validator.add_rule("mock_rule", MockRule)
        valid_value = await self.validator.validate_field("input1", 10, self.form)
        self.assertEqual(valid_value, 10)

        with self.assertRaises(FieldError):
            await self.validator.validate_field("input1", -5, self.form)

    async def test_validate_response(self):
        response = {"output": 20}
        validated_form = await self.validator.validate_response(self.form, response)
        self.assertEqual(validated_form.output, 20)

        response_str = "20"
        self.form.requested_fields = ["output"]
        validated_form = await self.validator.validate_response(self.form, response_str)
        self.assertEqual(validated_form.output, 20)

        with self.assertRaises(ValueError):
            await self.validator.validate_response(self.form, "invalid_response")

    async def test_validate_report(self):
        report = Report(assignment="a, b -> c", forms=[self.form])
        forms = [self.form, self.form2]
        validated_report = await self.validator.validate_report(report, forms)
        self.assertIsInstance(validated_report, Report)

    def test_add_rule(self):
        self.validator.add_rule("mock_rule", MockRule)
        self.assertIn("mock_rule", self.validator.active_rules)

    def test_remove_rule(self):
        self.validator.remove_rule("mock_rule")
        self.assertNotIn("mock_rule", self.validator.active_rules)

    def test_enable_rule(self):
        self.validator.enable_rule("mock_rule", enable=False)
        self.assertFalse(self.validator.active_rules["mock_rule"].enabled)
        self.validator.enable_rule("mock_rule", enable=True)
        self.assertTrue(self.validator.active_rules["mock_rule"].enabled)

    def test_disable_rule(self):
        self.validator.disable_rule("mock_rule")
        self.assertFalse(self.validator.active_rules["mock_rule"].enabled)

    def test_log_validation_attempt(self):
        result = {"output": 20}
        self.validator.log_validation_attempt(self.form, result)
        self.assertEqual(len(self.validator.validation_log), 1)

    def test_log_validation_error(self):
        self.validator.log_validation_error("input1", 10, "Error message")
        self.assertEqual(len(self.validator.validation_log), 1)

    def test_get_validation_summary(self):
        self.validator.log_validation_attempt(self.form, {"output": 20})
        self.validator.log_validation_error("input1", 10, "Error message")
        summary = self.validator.get_validation_summary()
        self.assertEqual(summary["total_attempts"], 2)
        self.assertEqual(len(summary["errors"]), 1)
        self.assertEqual(len(summary["successful_attempts"]), 1)


if __name__ == "__main__":
    unittest.main()



"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from enum import Enum
import asyncio
from typing import Any
from collections.abc import Coroutine

from lionagi.libs import SysUtil
from lionagi.core.collections.abc import Component


class WorkStatus(str, Enum):
    """Enum to represent different statuses of work."""

    PENDING = "PENDING"
    IN_PROGRESS = "IN_PROGRESS"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"


class Work(Component):
    """
    A class representing a unit of work.

    Attributes:
        status (WorkStatus): The current status of the work.
        result (Any): The result of the work, if completed.
        error (Any): Any error encountered during the work.
        async_task (Coroutine | None): The asynchronous task associated with the work.
        completion_timestamp (str | None): The timestamp when the work was completed.
        duration (float | None): The duration of the work.
    """

    status: WorkStatus = WorkStatus.PENDING
    result: Any = None
    error: Any = None
    async_task: Coroutine | None = None
    async_task_name: str | None = None
    completion_timestamp: str | None = None
    duration: float | None = None

    async def perform(self):
        """Perform the work and update the status, result, and duration."""
        try:
            result, duration = await self.async_task
            self.result = result
            self.status = WorkStatus.COMPLETED
            self.duration = duration
            del self.async_task
        except Exception as e:
            self.error = e
            self.status = WorkStatus.FAILED
        finally:
            self.completion_timestamp = SysUtil.get_timestamp(sep=None)[:-6]

    def __str__(self):
        return (
            f"Work(id={self.ln_id[:8]}.., status={self.status.value}, "
            f"created_at={self.timestamp[:-7]}, "
            f"completed_at={self.completion_timestamp[:-7]}, "
            f"duration={float(self.duration) if self.duration else 0:.04f} sec(s))"
        )



import unittest
from unittest.mock import AsyncMock, patch
from pydantic import ValidationError
from lionagi.core.collections.abc import Component, get_lion_id, LionIDable, Condition
from lionagi.core.generic.edge_condition import EdgeCondition
from lionagi.core.generic.edge import Edge


class TestEdge(unittest.TestCase):

    def setUp(self):
        self.node1 = Component()
        self.node2 = Component()
        self.edge = Edge(head=self.node1.ln_id, tail=self.node2.ln_id)

    def test_initialization(self):
        """Test initialization of the Edge class."""
        self.assertEqual(self.edge.head, self.node1.ln_id)
        self.assertEqual(self.edge.tail, self.node2.ln_id)
        self.assertIsNone(self.edge.condition)
        self.assertIsNone(self.edge.label)
        self.assertFalse(self.edge.bundle)

    async def test_check_condition_without_condition(self):
        """Test check_condition raises ValueError when condition is not set."""
        with self.assertRaises(ValueError):
            await self.edge.check_condition({})

    @patch("lionagi.core.models.edge.EdgeCondition.applies", new_callable=AsyncMock)
    async def test_check_condition_with_condition(self, mock_applies):
        """Test check_condition with a set condition."""
        mock_applies.return_value = True
        condition = EdgeCondition()
        self.edge.condition = condition
        result = await self.edge.check_condition({})
        self.assertTrue(result)
        mock_applies.assert_awaited_once()

    def test_validate_head_tail(self):
        """Test the head and tail validation."""
        try:
            valid_id = get_lion_id("valid_head")
        except Exception as e:
            if e.__class__.__name__ == "LionTypeError":
                return
        self.assertEqual(Edge._validate_head_tail(valid_id), valid_id)
        with self.assertRaises(ValidationError):
            Edge._validate_head_tail("invalid_id")

    def test_string_condition_none(self):
        """Test string_condition method when condition is None."""
        self.assertIsNone(self.edge.string_condition())

    def test_len(self):
        """Test the __len__ method."""
        self.assertEqual(len(self.edge), 1)

    def test_contains(self):
        """Test the __contains__ method."""
        self.assertIn(self.node1.ln_id, self.edge)
        self.assertIn(self.node2.ln_id, self.edge)
        fake_node = Component()
        self.assertNotIn(fake_node.ln_id, self.edge)


if __name__ == "__main__":
    unittest.main()



"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Any
from collections.abc import Mapping
from lionagi.libs.ln_convert import to_dict
from lionagi.libs import StringMatch, ParseUtil

from lionagi.core.rule.choice import ChoiceRule


class MappingRule(ChoiceRule):
    """
    Rule for validating that a value is a mapping (dictionary) with specific keys.

    Attributes:
        apply_type (str): The type of data to which the rule applies.
    """

    def __init__(self, apply_type="dict", **kwargs):
        super().__init__(apply_type=apply_type, **kwargs)

    async def validate(self, value: Any, *args, **kwargs) -> Any:
        """
        Validate that the value is a mapping with specific keys.

        Args:
            value (Any): The value to validate.
            *args: Additional arguments.
            **kwargs: Additional keyword arguments.

        Returns:
            Any: The validated value.

        Raises:
            ValueError: If the value is not a valid mapping or has incorrect keys.
        """
        if not isinstance(value, Mapping):
            raise ValueError("Invalid mapping field type.")

        if self.keys:
            if (keys := set(value.keys())) != set(self.keys):
                raise ValueError(
                    f"Invalid mapping keys. Current keys {[i for i in keys]} != {self.keys}"
                )
        return value

    async def perform_fix(self, value: Any, *args, **kwargs) -> Any:
        """
        Attempt to fix the value by converting it to a dict and validating its keys.

        Args:
            value (Any): The value to fix.
            *args: Additional arguments.
            **kwargs: Additional keyword arguments.

        Returns:
            Any: The fixed value.

        Raises:
            ValueError: If the value cannot be fixed.
        """
        if not isinstance(value, dict):
            try:
                value = to_dict(value)
            except Exception as e:
                raise ValueError("Invalid dict field type.") from e

        if self.keys:
            check_keys = set(value.keys())
            if check_keys != set(self.keys):
                try:
                    return StringMatch.force_validate_dict(value, keys=self.keys)
                except Exception as e:
                    raise ValueError("Invalid dict keys.") from e

        else:
            try:
                return ParseUtil.fuzzy_parse_json(value)
            except Exception as e:
                raise ValueError("Invalid dict keys.") from e

        return value



"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

"""
A class that manages asynchronous task processing with controlled concurrency.
"""

from typing import Any, Callable
import asyncio
from lionagi.libs import func_call


class AsyncQueue:
    """
    This class handles the enqueueing and processing of tasks with a limit on
    how many can run simultaneously, using an asyncio.Queue for task storage and
    an asyncio.Semaphore to limit concurrency.

    Attributes:
        queue (asyncio.Queue): The queue to store tasks.
        _stop_event (asyncio.Event): Event to signal processing should stop.
        max_concurrent_tasks (int): Maximum number of tasks processed concurrently.
        semaphore (asyncio.Semaphore): Controls concurrent access to task execution.
    """

    def __init__(self, max_concurrent_tasks=5):
        """
        Initializes the AsyncQueue with a concurrency limit.

        Args:
            max_concurrent_tasks (int): The maximum number of concurrent tasks
                                        allowed. Default is 5.
        """
        self.queue = asyncio.Queue()
        self._stop_event = asyncio.Event()
        self.max_concurrent_tasks = max_concurrent_tasks
        self.semaphore = asyncio.Semaphore(max_concurrent_tasks)

    async def enqueue(self, input_: Any) -> None:
        """
        Enqueues an item to be processed asynchronously.

        Args:
            input_ (Any): The item to be enqueued.
        """
        await self.queue.put(input_)

    async def dequeue(self) -> Any:
        """
        Dequeues an item for processing.

        Returns:
            Any: The dequeued item.
        """
        return await self.queue.get()

    async def join(self) -> None:
        """Waits for all items in the queue to be processed."""
        await self.queue.join()

    async def stop(self) -> None:
        """Signals the queue to stop processing new items."""
        self._stop_event.set()

    def stopped(self) -> bool:
        """
        Checks if the stop signal has been issued.

        Returns:
            bool: True if the queue has been stopped, otherwise False.
        """
        return self._stop_event.is_set()

    async def process_requests(self, func: Callable, retry_kwargs: dict = {}) -> None:
        """
        Processes tasks from the queue using the provided function with retries.

        This method continuously processes tasks from the queue using the specified
        function until a stop event is triggered. Handles concurrency using a
        semaphore and manages task completion.

        Args:
            func (Callable): The function to process each task.
            retry_kwargs (dict): Keyword arguments for retry behavior. Default is
                                 an empty dictionary.
        """
        tasks = set()
        while not self.stopped():
            if len(tasks) >= self.max_concurrent_tasks:
                _, done = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)
                tasks.difference_update(done)

            async with self.semaphore:
                input_ = await self.dequeue()
                if input_ is None:
                    await self.stop()
                    break
                task = asyncio.create_task(
                    func_call.rcall(func, input_, **retry_kwargs)
                )
                tasks.add(task)

        if tasks:
            await asyncio.wait(tasks)



"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Any
from lionagi.libs.ln_convert import is_same_dtype, to_df
from lionagi.core.collections.abc import Field
from lionagi.core.collections import (
    pile,
    progression,
    Pile,
    Progression,
    iModel,
    Exchange,
)
from lionagi.core.generic.node import Node
from lionagi.core.action import Tool, ToolManager
from lionagi.core.mail import Mail, Package
from lionagi.core.message import (
    create_message,
    System,
    Instruction,
    AssistantResponse,
    ActionRequest,
    ActionResponse,
    RoledMessage,
)

from lionagi.core.session.directive_mixin import DirectiveMixin


class Branch(Node, DirectiveMixin):
    """
    A class representing a branch in a messaging system.

    Attributes:
        messages (Pile): A pile of messages.
        progress (Progression): A progression of messages.
        tool_manager (ToolManager): A manager for handling tools.
        system (System): The system associated with the branch.
        user (str): The user associated with the branch.
        mailbox (Exchange): An exchange for managing mail.
        imodel (iModel): The model associated with the branch.
    """

    messages: Pile = Field(None)
    progress: Progression = Field(None)
    tool_manager: ToolManager = Field(None)
    system: System = Field(None)
    user: str = Field(None)
    mailbox: Exchange[Mail] = Field(None)
    imodel: iModel = Field(None)

    def __init__(
        self,
        system: System | None = None,
        system_sender: str | None = None,
        user: str | None = None,
        messages: Pile = None,
        progress: Progression = None,
        tool_manager: ToolManager = None,
        tools: Any = None,
        imodel=None,
    ):
        """
        Initializes a new instance of the Branch class.

        Args:
            system (System, optional): The system associated with the branch.
            system_sender (str, optional): The sender of the system message.
            user (str, optional): The user associated with the branch.
            messages (Pile, optional): A pile of messages.
            progress (Progression, optional): A progression of messages.
            tool_manager (ToolManager, optional): A manager for handling tools.
            tools (Any, optional): Tools to be registered with the tool manager.
            imodel (iModel, optional): The model associated with the branch.
        """
        super().__init__()
        self.system = None

        self.user = user or "user"
        self.messages = messages or pile({})
        self.progress = progress or progression()
        self.tool_manager = tool_manager or ToolManager()
        self.mailbox = Exchange()
        self.imodel = imodel or iModel()
        if tools:
            self.tool_manager.register_tools(tools)
        self.set_system(system=system, sender=system_sender)
        # system = system or "You are a helpful assistant, let's think step by step"
        # self.add_message(system=system, sender=system_sender)

    def set_system(self, system=None, sender=None) -> None:
        """
        Sets the system message.

        Args:
            system (System): The system message to set.
            sender (str, optional): The sender of the system message.
        """
        system = system or "You are a helpful assistant."
        if len(self.progress) == 0:
            self.add_message(system=system, sender=sender)
        else:
            _msg = System(system=system, sender=sender)
            _msg.recipient = self.ln_id
            self._remove_system()
            self.system = _msg

    def add_message(
        self,
        *,
        system=None,  # system node - JSON serializable
        instruction=None,  # Instruction node - JSON serializable
        context=None,  # JSON serializable
        assistant_response=None,  # JSON
        function=None,
        arguments=None,
        func_outputs=None,
        action_request=None,  # ActionRequest node
        action_response=None,  # ActionResponse node
        images=None,
        sender=None,  # str
        recipient=None,  # str
        requested_fields=None,  # dict[str, str]
        metadata: dict | None = None,  # extra metadata
        **kwargs,  # additional context fields
    ) -> bool:
        """
        Adds a message to the branch.

        Args:
            system (Any, optional): The system node (JSON serializable).
            instruction (Any, optional): The instruction node (JSON serializable).
            context (Any, optional): Additional context (JSON serializable).
            assistant_response (Any, optional): The assistant's response (JSON serializable).
            function (Any, optional): The function associated with the message.
            arguments (Any, optional): The arguments for the function.
            func_outputs (Any, optional): The outputs of the function.
            action_request (Any, optional): The action request node.
            action_response (Any, optional): The action response node.
            sender (str, optional): The sender of the message.
            recipient (str, optional): The recipient of the message.
            requested_fields (dict[str, str], optional): Requested fields for the message.
            metadata (dict, optional): Extra metadata for the message.
            **kwargs: Additional context fields.

        Returns:
            bool: True if the message was successfully added, else False.
        """
        if assistant_response:
            sender = self.ln_id

        _msg = create_message(
            system=system,
            instruction=instruction,
            context=context,
            assistant_response=assistant_response,
            function=function,
            arguments=arguments,
            func_outputs=func_outputs,
            action_request=action_request,
            action_response=action_response,
            sender=sender,
            images=images,
            recipient=recipient,
            requested_fields=requested_fields,
            **kwargs,
        )

        if isinstance(_msg, System):
            _msg.recipient = self.ln_id  # the branch itself, system is to the branch
            self._remove_system()
            self.system = _msg

        if isinstance(_msg, Instruction):
            _msg.sender = sender or self.user
            _msg.recipient = recipient or self.ln_id

        if isinstance(_msg, AssistantResponse):
            _msg.sender = sender or self.ln_id
            _msg.recipient = recipient or "user"

        if isinstance(_msg, ActionRequest):
            _msg.sender = sender or self.ln_id
            _msg.recipient = recipient or "N/A"

        if isinstance(_msg, ActionResponse):
            _msg.sender = sender or "N/A"
            _msg.recipient = recipient or self.ln_id

        if metadata:
            _msg._meta_insert(["extra"], metadata)

        return self.messages.include(_msg) and self.progress.include(_msg)

    def to_chat_messages(self) -> list[dict[str, Any]]:
        """
        Converts the messages to chat message format.

        Returns:
            list[dict[str, Any]]: A list of chat messages.
        """
        return [self.messages[j].chat_msg for j in self.progress]

    def _remove_system(self) -> None:
        """
        Removes the system message from the branch.
        """
        self.messages.exclude(self.system)
        self.progress.exclude(self.system)
        self.system = None

    def clear(self) -> None:
        """
        Clears all messages and progression in the branch.
        """
        self.messages.clear()
        self.progress.clear()

    @property
    def has_tools(self) -> bool:
        """
        Checks if the branch has tools.

        Returns:
            bool: True if the branch has tools, else False.
        """
        return self.tool_manager.registry != {}

    def register_tools(self, tools) -> None:
        """
        Registers tools with the tool manager.

        Args:
            tools (Any): The tools to register.
        """
        self.tool_manager.register_tools(tools=tools)

    def delete_tools(self, tools, verbose: bool = True) -> bool:
        """
        Deletes tools from the tool manager.

        Args:
            tools (Any): The tools to delete.
            verbose (bool, optional): Whether to print deletion status.

        Returns:
            bool: True if tools were successfully deleted, else False.
        """
        if not isinstance(tools, list):
            tools = [tools]
        if is_same_dtype(tools, str):
            for act_ in tools:
                if act_ in self.tool_manager.registry:
                    self.tool_manager.registry.pop(act_)
            if verbose:
                print("tools successfully deleted")
            return True
        elif is_same_dtype(tools, Tool):
            for act_ in tools:
                if act_.schema_["function"]["name"] in self.tool_manager.registry:
                    self.tool_manager.registry.pop(act_.schema_["function"]["name"])
            if verbose:
                print("tools successfully deleted")
            return True
        if verbose:
            print("tools deletion failed")
        return False

    def update_last_instruction_meta(self, meta):
        """
        Updates metadata of the last instruction.

        Args:
            meta (dict): The metadata to update.
        """

        for i in reversed(self.progress):
            if isinstance(self.messages[i], Instruction):
                self.messages[i]._meta_insert(["extra"], meta)
                return

    @property
    def last_response(self):
        for i in reversed(self.progress):
            if isinstance(self.messages[i], AssistantResponse):
                return self.messages[i]

    @property
    def assistant_responses(self):
        return pile(
            [
                self.messages[i]
                for i in self.progress
                if isinstance(self.messages[i], AssistantResponse)
            ]
        )

    def to_df(self) -> Any:
        """
        Converts the messages to a DataFrame.

        Returns:
            Any: A DataFrame representation of the messages.
        """
        fields = [
            "ln_id",
            "message_type",
            "timestamp",
            "role",
            "content",
            "metadata",
            "sender",
            "recipient",
        ]
        dicts_ = []
        for i in self.progress:
            _d = {}
            for j in fields:
                _d.update({j: getattr(self.messages[i], j, None)})
                _d["message_type"] = self.messages[i].class_name
            dicts_.append(_d)

        return to_df(dicts_)

    def _is_invoked(self) -> bool:
        """
        Checks if the last message is an ActionResponse.

        Returns:
            bool: True if the last message is an ActionResponse, else False.
        """
        return isinstance(self.messages[-1], ActionResponse)

    def send(
        self, recipient: str, category: str, package: Any, request_source: str = None
    ) -> None:
        """
        Sends a mail to a recipient.

        Args:
            recipient (str): The ID of the recipient.
            category (str): The category of the mail.
            package (Any): The package to send in the mail.
            request_source (str): The source of the request.
        """
        pack = Package(
            category=category, package=package, request_source=request_source
        )
        mail = Mail(
            sender=self.ln_id,
            recipient=recipient,
            package=pack,
        )
        self.mailbox.include(mail, "out")

    def receive(
        self,
        sender: str,
        message: bool = True,
        tool: bool = True,
        imodel: bool = True,
    ) -> None:
        """
        Receives mail from a sender.

        Args:
            sender (str): The ID of the sender.
            message (bool, optional): Whether to process message mails. Defaults to True.
            tool (bool, optional): Whether to process tool mails. Defaults to True.
            imodel (bool, optional): Whether to process imodel mails. Defaults to True.

        Raises:
            ValueError: If the sender does not exist or the mail category is invalid.
        """
        skipped_requests = progression()
        if sender not in self.mailbox.pending_ins.keys():
            raise ValueError(f"No package from {sender}")
        while self.mailbox.pending_ins[sender].size() > 0:
            mail_id = self.mailbox.pending_ins[sender].popleft()
            mail: Mail = self.mailbox.pile[mail_id]

            if mail.category == "message" and message:
                if not isinstance(mail.package.package, RoledMessage):
                    raise ValueError("Invalid message format")
                new_message = mail.package.package.clone()
                new_message.sender = mail.sender
                new_message.recipient = self.ln_id
                self.messages.include(new_message)
                self.progress.include(new_message)
                self.mailbox.pile.pop(mail_id)

            elif mail.category == "tool" and tool:
                if not isinstance(mail.package.package, Tool):
                    raise ValueError("Invalid tools format")
                self.tool_manager.register_tools(mail.package.package)
                self.mailbox.pile.pop(mail_id)

            elif mail.category == "imodel" and imodel:
                if not isinstance(mail.package.package, iModel):
                    raise ValueError("Invalid iModel format")
                self.imodel = mail.package.package
                self.mailbox.pile.pop(mail_id)

            else:
                skipped_requests.append(mail)

        self.mailbox.pending_ins[sender] = skipped_requests

        if self.mailbox.pending_ins[sender].size() == 0:
            self.mailbox.pending_ins.pop(sender)

    def receive_all(self) -> None:
        """
        Receives mail from all senders.
        """
        for key in list(self.mailbox.pending_ins.keys()):
            self.receive(key)



import unittest
from lionagi import Node, pile, progression, flow


class TestFlow(unittest.TestCase):

    def setUp(self):
        # Setup for creating initial nodes, piles, and progressions
        self.nodes = [Node(content=i) for i in range(10)]
        self.pile1 = pile(self.nodes)
        self.prog1 = progression(self.nodes[:5], name="left")
        self.prog2 = progression(self.nodes[5:], name="right")
        self.flow = flow([self.prog1, self.prog2])

    def test_initialization(self):
        self.assertEqual(len(self.flow.sequences), 2)
        self.assertIn("left", self.flow.registry)
        self.assertIn("right", self.flow.registry)

    def test_append_to_flow(self):
        new_node = Node(content="new_node")
        self.flow.append(new_node, "left")
        self.assertIn(new_node.ln_id, self.flow.get("left"))

    def test_exclude_from_flow(self):
        node_to_remove = self.nodes[0]
        self.flow.exclude("left", node_to_remove)
        self.assertNotIn(node_to_remove.ln_id, self.flow.get("left"))

    def test_get_sequence(self):
        seq = self.flow.get("left")
        self.assertEqual(seq, self.prog1)
        seq_default = self.flow.get("nonexistent", default=None)
        self.assertIsNone(seq_default)

    def test_register_sequence(self):
        new_prog = progression(self.nodes[2:4], name="middle")
        self.flow.register(new_prog)
        self.assertIn("middle", self.flow.registry)

    def test_popleft(self):
        first_node_id = self.prog1[0]
        removed_node_id = self.flow.popleft("left")
        self.assertEqual(first_node_id, removed_node_id)
        self.assertNotIn(removed_node_id, self.flow.get("left"))

    def test_shape_and_size(self):
        shape = self.flow.shape()
        self.assertEqual(shape["left"], 5)
        self.assertEqual(shape["right"], 5)
        size = self.flow.size()
        self.assertEqual(size, 10)

    def test_clear(self):
        self.flow.clear()
        self.assertEqual(len(self.flow.sequences), 0)
        self.assertEqual(len(self.flow.registry), 0)

    def test_iteration(self):
        items = list(self.flow)
        self.assertEqual(len(items), 2)

    def test_to_df(self):
        df = self.flow.to_df()
        self.assertEqual(df.shape[0], 2)
        self.assertEqual(df.columns.tolist(), ["order", "name"])

    def test_inclusion_check(self):
        self.assertTrue(self.prog1 in self.flow)
        self.assertTrue(self.prog2 in self.flow)
        try:
            a = Node(content="nonexistent") in self.flow
        except TypeError:
            pass

    # Additional Tests for Edge Cases
    def test_empty_flow_initialization(self):
        empty_flow = flow()
        self.assertEqual(len(empty_flow.sequences), 0)
        self.assertEqual(len(empty_flow.registry), 0)

    def test_append_to_nonexistent_sequence(self):
        new_node = Node(content="new_node")
        self.flow.append(new_node, "nonexistent")
        self.assertIn(new_node.ln_id, self.flow.get("nonexistent"))

    def test_exclude_nonexistent_item(self):
        non_existent_node = Node(content="nonexistent")
        result = self.flow.exclude("left", non_existent_node)
        self.assertTrue(result)

    def test_exclude_nonexistent_sequence(self):
        result = self.flow.exclude("nonexistent")
        self.assertFalse(result)

    def test_get_nonexistent_sequence(self):
        try:
            self.flow.get("nonexistent")
        except Exception as e:
            if e.__class__.__name__ == "LionTypeError":
                return
            raise AssertionError("LionTypeError not raised")

    def test_register_existing_sequence_name(self):
        with self.assertRaises(ValueError):
            self.flow.register(self.prog1, name="right")

    def test_popleft_nonexistent_sequence(self):
        try:
            self.flow.popleft("nonexistent")
        except Exception as e:
            if e.__class__.__name__ == "LionTypeError":
                return
            raise AssertionError("LionTypeError not raised")

    def test_append_multiple_items(self):
        new_nodes = [Node(content="new_node1"), Node(content="new_node2")]
        for new_node in new_nodes:
            self.flow.append(new_node, "left")
        for new_node in new_nodes:
            self.assertIn(new_node.ln_id, self.flow.get("left"))

    def test_remove_from_all_sequences(self):
        node_to_remove = self.nodes[0]
        self.flow.append(node_to_remove, "right")
        self.flow.remove(node_to_remove)
        self.assertNotIn(node_to_remove.ln_id, self.flow.get("right"))

    def test_shape_empty_flow(self):
        empty_flow = flow()
        self.assertEqual(empty_flow.shape(), {})

    def test_size_empty_flow(self):
        empty_flow = flow()
        self.assertEqual(empty_flow.size(), 0)

    def test_clear_empty_flow(self):
        empty_flow = flow()
        empty_flow.clear()
        self.assertEqual(len(empty_flow.sequences), 0)
        self.assertEqual(len(empty_flow.registry), 0)


if __name__ == "__main__":
    unittest.main()



"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import ast
import operator


class ASTEvaluator:
    """
    Safely evaluates expressions using AST parsing to prevent unsafe operations.
    """

    def __init__(self):
        self.allowed_operators = {
            ast.Eq: operator.eq,
            ast.NotEq: operator.ne,
            ast.Lt: operator.lt,
            ast.LtE: operator.le,
            ast.Gt: operator.gt,
            ast.GtE: operator.ge,
            # Additional operators can be added here as needed
        }

    def evaluate(self, expression, context):
        """
        Evaluate a condition expression within a given context using AST parsing.
        """
        try:
            tree = ast.parse(expression, mode="eval")
            return self._evaluate_node(tree.body, context)
        except Exception as e:
            raise ValueError(f"Failed to evaluate expression: {expression}. Error: {e}")

    def _evaluate_node(self, node, context):
        if isinstance(node, ast.Compare):
            left = self._evaluate_node(node.left, context)
            for operation, comparator in zip(node.ops, node.comparators):
                op_func = self.allowed_operators.get(type(operation))
                if not op_func:
                    raise ValueError(
                        f"Operation {type(operation).__name__} is not allowed."
                    )
                right = self._evaluate_node(comparator, context)
                if not op_func(left, right):
                    return False
            return True
        elif isinstance(node, ast.Name):
            return context.get(node.id)
        elif isinstance(node, ast.Constant):
            return node.n
        else:
            raise ValueError(
                "Unsupported AST node type encountered in condition evaluation."
            )


class ASTEvaluationEngine:
    """
    Executes scripts safely using the SafeEvaluator for expression evaluation.
    """

    def __init__(self):
        self.variables = {}
        self.safe_evaluator = ASTEvaluator()
        self.functions = {
            "processData": self.process_data,
        }

    def process_data(self, data):
        # Example placeholder function for data processing
        return data * 2

    def _evaluate_expression(self, expression):
        """
        Evaluates expressions within scripts using SafeEvaluator.
        """
        # Here, 'self.variables' serves as the context for the evaluation
        return self.safe_evaluator.evaluate(expression, self.variables)

    def _assign_variable(self, var_name, value):
        """
        Assigns a value to a variable within the script's context.
        """
        self.variables[var_name] = value

    def _execute_function(self, func_name, arg):
        """
        Executes a predefined function with the given argument.
        """
        if func_name in self.functions:
            function = self.functions[func_name]
            return function(arg)
        else:
            raise ValueError(f"Function '{func_name}' is not defined.")

    def execute(self, script):
        """
        Parses and executes a script, handling variable assignments and function calls.
        """
        tree = ast.parse(script, mode="exec")
        for stmt in tree.body:
            if isinstance(stmt, ast.Assign):
                var_name = stmt.targets[
                    0
                ].id  # Assumes single target assignment for simplicity
                # Convert the AST node back to a string for evaluation
                value_expr = ast.unparse(stmt.value)
                value = self._evaluate_expression(value_expr)
                self._assign_variable(var_name, value)
            elif isinstance(stmt, ast.Expr) and isinstance(stmt.value, ast.Call):
                func_name = stmt.value.func.id
                arg_expr = ast.unparse(stmt.value.args[0])
                arg = self._evaluate_expression(arg_expr)
                self._execute_function(func_name, arg)
            else:
                raise ValueError(
                    "Unsupported statement type encountered in script execution."
                )



# from lionagi.libs.ln_convert import is_same_dtype
# from collections.abc import Mapping, Generator


# def validate_keys(keys):
#     """
#     choices can be provided from various sources:
#     - mapping such as dict, their keys will be used as choices
#     - iterables including list, tuple, set, generator, enum, etc.
#     - strings, comma separated values
#     """

#     try:
#         if isinstance(keys, Mapping):
#             keys = list(keys.keys())

#         elif isinstance(keys, (list, tuple, set, Generator)):
#             keys = set(keys)

#         elif isinstance(keys, str):
#             if "," in keys:
#                 keys = list({i.strip() for i in keys.split(",")})
#             else:
#                 keys = [keys.strip()]

#         else:
#             keys = [i.value for i in keys]

#     except Exception as e:
#         raise ValueError(f"invalid choices {keys}") from e

#     if not is_same_dtype(keys):
#         raise ValueError(f"choices must be of the same type, got {keys}")

#     return keys



"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from lionagi.core.collections import (
    Pile,
    Progression,
    progression,
    pile,
    iModel,
)
from lionagi.core.message import System
from typing import Any
from lionagi.core.action.tool_manager import ToolManager

from lionagi.libs import SysUtil
from lionagi.core.session.branch import Branch
from lionagi.core.collections import pile, Pile, Exchange
from lionagi.core.collections.abc import get_lion_id
from lionagi.core.collections.util import to_list_type
from lionagi.core.mail.mail_manager import MailManager


class Session:
    """
    A session for managing branches, mail transfer, and interactions with a model.

    Attributes:
        ln_id (str): The unique identifier for the session.
        timestamp (str): The timestamp when the session was created.
        system (System): The default system message for the session.
        system_sender (str): The sender of the system message.
        branches (Pile[Branch]): The pile of branches in the session.
        mail_transfer (Exchange): The exchange for managing mail transfer.
        mail_manager (MailManager): The manager for handling mail.
        imodel (iModel): The model associated with the session.
        user (str): The user associated with the session.
        default_branch (Branch): The default branch of the session.
    """

    def __init__(
        self,
        system=None,  # the default system message for the session
        branches: Any | None = None,
        system_sender: str | None = None,
        user: str | None = None,
        imodel=None,
        tools=None,
    ):
        self.ln_id = SysUtil.create_id()
        self.timestamp = SysUtil.get_timestamp(sep=None)[:-6]
        system = system or "You are a helpful assistant, let's think step by step"
        self.system = System(system=system, sender=system_sender)
        self.system_sender = system_sender
        self.branches: Pile[Branch] = self._validate_branches(branches)
        self.mail_transfer = Exchange()
        self.mail_manager = MailManager([self.mail_transfer])
        self.imodel = imodel or iModel()
        self.user = user
        self.default_branch = None
        if self.branches.size() == 0:
            self.new_branch(system=self.system.clone())
        else:
            self.default_branch = self.branches[0]
        if tools:
            self.default_branch.tool_manager.register_tools(tools)

    def _validate_branches(self, value):
        """
        Validates and converts the branches input to a Pile of Branch objects.

        Args:
            value (Any): The input value to validate and convert.

        Returns:
            Pile[Branch]: A pile of validated branches.

        Raises:
            ValueError: If the input value contains non-Branch objects.
        """
        if isinstance(value, Pile):
            for branch in value:
                if not isinstance(branch, Branch):
                    raise ValueError("The branches pile contains non-Branch object")
            return value
        else:
            try:
                value = pile(items=value, item_type=Branch)
                return value
            except Exception as e:
                raise ValueError(f"Invalid branches value. Error:{e}")

    # ---- branch manipulation ---- #
    def new_branch(
        self,
        system: System | None = None,
        system_sender: str | None = None,
        user: str | None = None,
        messages: Pile = None,
        progress: Progression = None,
        tool_manager: ToolManager = None,
        tools: Any = None,
        imodel=None,
    ):
        """
        Creates a new branch and adds it to the session.

        Args:
            system (System, optional): The system message for the branch.
            system_sender (str, optional): The sender of the system message.
            user (str, optional): The user associated with the branch.
            messages (Pile, optional): The pile of messages for the branch.
            progress (Progression, optional): The progression of messages.
            tool_manager (ToolManager, optional): The tool manager for the branch.
            tools (Any, optional): The tools to register with the tool manager.
            imodel (iModel, optional): The model associated with the branch.

        Returns:
            Branch: The created branch.
        """
        if system is None:
            system = self.system.clone()
            system.sender = self.ln_id
            system_sender = self.ln_id
        branch = Branch(
            system=system,
            system_sender=system_sender,
            user=user,
            messages=messages,
            progress=progress,
            tool_manager=tool_manager,
            tools=tools,
            imodel=imodel or self.imodel,
        )
        self.branches.append(branch)
        self.mail_manager.add_sources(branch)
        if self.default_branch is None:
            self.default_branch = branch
        return branch

    def delete_branch(self, branch):
        """
        Deletes a branch from the session.

        Args:
            branch (Branch | str): The branch or its ID to delete.
        """
        branch_id = get_lion_id(branch)
        self.branches.pop(branch_id)
        self.mail_manager.delete_source(branch_id)

        if self.default_branch == branch:
            if self.branches.size() == 0:
                self.default_branch = None
            else:
                self.default_branch = self.branches[0]

    def split_branch(self, branch):
        """
        Splits a branch, creating a new branch with the same messages and tools.

        Args:
            branch (Branch | str): The branch or its ID to split.

        Returns:
            Branch: The newly created branch.
        """
        branch = self.branches[branch]
        system = branch.system.clone() if branch.system else None
        if system:
            system.sender = branch.ln_id
        progress = progression()
        messages = pile()

        for id_ in branch.progress:
            clone_message = branch.messages[id_].clone()
            progress.append(clone_message.ln_id)
            messages.append(clone_message)

        tools = (
            list(branch.tool_manager.registry.values())
            if branch.tool_manager.registry
            else None
        )
        branch_clone = Branch(
            system=system,
            system_sender=branch.ln_id,
            user=branch.user,
            progress=progress,
            messages=messages,
            tools=tools,
        )
        for message in branch_clone.messages:
            message.sender = branch.ln_id
            message.recipient = branch_clone.ln_id
        self.branches.append(branch_clone)
        self.mail_manager.add_sources(branch_clone)
        return branch_clone

    def change_default_branch(self, branch):
        """
        Changes the default branch of the session.

        Args:
            branch (Branch | str): The branch or its ID to set as the default.
        """
        branch = self.branches[branch]
        self.default_branch = branch

    def collect(self, from_: Branch | str | Pile[Branch] | None = None):
        """
        Collects mail from specified branches.

        Args:
            from_ (Branch | str | Pile[Branch], optional): The branches to collect mail from.
                If None, collects mail from all branches.
        """
        if from_ is None:
            self.mail_manager.collect_all()
        else:
            try:
                sources = to_list_type(from_)
                for source in sources:
                    self.mail_manager.collect(get_lion_id(source))
            except Exception as e:
                raise ValueError(f"Failed to collect mail. Error: {e}")

    def send(self, to_: Branch | str | Pile[Branch] | None = None):
        """
        Sends mail to specified branches.

        Args:
            to_ (Branch | str | Pile[Branch], optional): The branches to send mail to.
                If None, sends mail to all branches.
        """
        if to_ is None:
            self.mail_manager.send_all()
        else:
            try:
                sources = to_list_type(to_)
                for source in sources:
                    self.mail_manager.send(get_lion_id(source))
            except Exception as e:
                raise ValueError(f"Failed to send mail. Error: {e}")

    def collect_send_all(self, receive_all=False):
        """
        Collects and sends mail for all branches, optionally receiving all mail.

        Args:
            receive_all (bool, optional): Whether to receive all mail for all branches.
        """
        self.collect()
        self.send()
        if receive_all:
            for branch in self.branches:
                branch.receive_all()

    async def chat(self, *args, branch=None, **kwargs):
        """
        Initiates a chat interaction with a branch.

        Args:
            *args: Positional arguments to pass to the chat method.
            branch (Branch, optional): The branch to chat with. Defaults to the default branch.
            **kwargs: Keyword arguments to pass to the chat method.

        Returns:
            Any: The result of the chat interaction.

        Raises:
            ValueError: If the specified branch is not found in the session branches.
        """
        if branch is None:
            branch = self.default_branch
        if branch not in self.branches:
            raise ValueError("Branch not found in session branches")
        return await self.branches[branch].chat(*args, **kwargs)

    async def direct(self, *args, branch=None, **kwargs):
        """
        Initiates a direct interaction with a branch.

        Args:
            *args: Positional arguments to pass to the direct method.
            branch (Branch, optional): The branch to interact with. Defaults to the default branch.
            **kwargs: Keyword arguments to pass to the direct method.

        Returns:
            Any: The result of the direct interaction.

        Raises:
            ValueError: If the specified branch is not found in the session branches.
        """
        if branch is None:
            branch = self.default_branch
        if branch not in self.branches:
            raise ValueError("Branch not found in session branches")
        return await self.branches[branch].direct(*args, **kwargs)



"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
from .version import __version__
from dotenv import load_dotenv

from lionagi.libs.ln_convert import to_list, to_dict, to_df, to_readable_dict
from lionagi.libs.ln_func_call import alcall, bcall, lcall, CallDecorator as cd, tcall
from lionagi.core.collections.abc import Field
from lionagi.core.collections import progression, flow, pile, iModel
from lionagi.core.generic import Node, Graph, Tree, Edge
from lionagi.core.action import func_to_tool
from lionagi.core.report import Form, Report
from lionagi.core.session.branch import Branch
from lionagi.core.session.session import Session
from lionagi.core.work.worker import work, Worker, worklink
from lionagi.integrations.provider.services import Services
from lionagi.integrations.chunker.chunk import chunk
from lionagi.integrations.loader.load import load
import lionagi.core.director.direct as direct


__all__ = [
    "Field",
    "progression",
    "flow",
    "pile",
    "iModel",
    "work",
    "worklink",
    "Worker",
    "Branch",
    "Session",
    "Form",
    "Report",
    "Services",
    "direct",
    "Node",
    "Graph",
    "Tree",
    "Edge",
    "chunk",
    "load",
    "func_to_tool",
    "cd",
    "alcall",
    "bcall",
    "to_list",
    "to_dict",
    "lcall",
    "to_df",
    "tcall",
    "to_readable_dict",
]


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
load_dotenv()



"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

"""
This module introduces the Report class, an extension of the BaseForm class
designed to manage and synchronize a collection of Form instances based on
specific assignments. The Report class handles the creation and updating of
forms, ensuring each is properly configured according to the report's
requirements.
"""

from typing import Any, Type
from lionagi.core.collections.abc import Field
from lionagi.core.collections import Pile, pile
from lionagi.core.report.util import get_input_output_fields
from lionagi.core.report.base import BaseForm
from lionagi.core.report.form import Form


class Report(BaseForm):
    """
    Extends BaseForm to handle a collection of Form instances based on specific
    assignments, managing a pile of forms and ensuring synchronization and
    proper configuration.
    """

    forms: Pile[Form] = Field(
        None,
        description="A pile of forms related to the report.",
    )

    form_assignments: list = Field(
        [],
        description="assignment for the report",
        examples=[["a, b -> c", "a -> e", "b -> f", "c -> g", "e, f, g -> h"]],
    )

    form_template: Type[Form] = Field(
        Form, description="The template for the forms in the report."
    )

    def __init__(self, **kwargs):
        """
        Initializes the Report with input and requested fields based on the
        report's assignment, creating forms dynamically from provided assignments.
        """
        super().__init__(**kwargs)
        self.input_fields, self.requested_fields = get_input_output_fields(
            self.assignment
        )

        # if assignments is not provided, set it to report assignment
        if self.form_assignments == []:
            self.form_assignments.append(self.assignment)

        # create forms
        self.forms = pile(
            [self.form_template(assignment=i) for i in self.form_assignments],
            [Form, BaseForm, Report],
        )

        # Add undeclared fields to report with None values
        for v in self.forms:
            for _field in list(v.work_fields.keys()):
                if _field not in self._all_fields:
                    field_obj = v._all_fields[_field]
                    self._add_field(_field, value=None, field_obj=field_obj)

        # Synchronize fields between report and forms
        for k, v in self._all_fields.items():
            if getattr(self, k, None) is not None:
                for _form in self.forms:
                    if k in _form.work_fields:
                        _form.fill(**{k: getattr(self, k)})

    @property
    def work_fields(self) -> dict[str, Any]:
        all_fields = {}
        for form in self.forms.values():
            for k, v in form.work_fields.items():
                if k not in all_fields:
                    all_fields[k] = v
        return all_fields

    def fill(self, form: Form | list[Form] | dict[Form] = None, strict=True, **kwargs):
        if self.filled:
            if strict:
                raise ValueError("Form is filled, cannot be worked on again")

        # gather all unique valid fields from input form,
        # kwargs and self workfields data
        all_fields = self._get_all_fields(form, **kwargs)

        # if there are information in the forms that are not in the report,
        # add them to the report
        for k, v in all_fields.items():
            if k in self.work_fields and getattr(self, k, None) is None:
                setattr(self, k, v)

        # if there are information in the report that are not in the forms,
        # add them to the forms
        for _form in self.forms:
            for k, v in _form.work_fields.items():
                _kwargs = {}
                if v is None and (a := getattr(self, k, None)) is not None:
                    _kwargs[k] = a
                _form.fill(**_kwargs)

    def is_workable(self) -> bool:
        """
        Checks if the report is ready for processing, ensuring all necessary fields
        are filled and output fields are unique across forms.

        Returns:
            bool: True if the report is workable, otherwise raises ValueError.
        """
        if self.filled:
            raise ValueError("Form is already filled, cannot be worked on again")

        for i in self.input_fields:
            if not getattr(self, i, None):
                raise ValueError(f"Required field {i} is not provided")

        # this is the required fields from report's own assignment
        fields = self.input_fields
        fields.extend(self.requested_fields)

        # if the report's own assignment is not in the forms, return False
        for f in fields:
            if f not in self.work_fields:
                raise ValueError(f"Field {f} is not in the forms")

        # get all the output fields from all the forms
        outs = []
        for form in self.forms.values():
            outs.extend(form.requested_fields)

        # all output fields should be unique, not a single output field should be
        # calculated by more than one form
        if len(outs) != len(set(outs)):
            raise ValueError("Output fields are not unique")

        return True

    def next_forms(self) -> Pile[Form]:
        """
        Returns a pile of workable forms based on current form states within the report.

        Returns:
            Pile[Form]: A pile of workable forms or None if there are no workable forms.
        """
        a = [i for i in self.forms if i.workable]
        return pile(a, Form) if len(a) > 0 else None



#TODO
# import unittest
# from unittest.mock import MagicMock, patch
# from lionagi.core.generic import BaseNode, Edge
# from lionagi.new.schema.todo.graph import Graph
#
#
# class TestGraph(unittest.TestCase):
#     def setUp(self):
#         self.graph = Graph()
#         self.node1 = BaseNode(id_="node1", content="Node 1")
#         self.node2 = BaseNode(id_="node2", content="Node 2")
#         self.edge1 = Edge(
#             id_="edge1", source_node_id="node1", target_node_id="node2", label="Edge 1"
#         )
#         self.graph.structure_nodes = {"node1": self.node1, "node2": self.node2}
#         self.graph.structure_edges = {"edge1": self.edge1}
#
#     @patch("lionagi.libs.SysUtil.check_import")
#     def test_to_networkx_success(self, mock_check_import):
#         mock_check_import.return_value = None
#         with patch("networkx.DiGraph") as mock_digraph:
#             mock_graph = MagicMock()
#             mock_digraph.return_value = mock_graph
#
#             result = self.graph.to_networkx()
#
#             self.assertEqual(result, mock_graph)
#
#     @patch("lionagi.libs.SysUtil.check_import")
#     def test_to_networkx_empty_graph(self, mock_check_import):
#         mock_check_import.return_value = None
#         with patch("networkx.DiGraph") as mock_digraph:
#             mock_graph = MagicMock()
#             mock_digraph.return_value = mock_graph
#
#             self.graph.structure_nodes = {}
#             self.graph.structure_edges = {}
#             result = self.graph.to_networkx()
#
#             self.assertEqual(result, mock_graph)
#             mock_check_import.assert_called_once_with("networkx")
#             mock_digraph.assert_called_once()
#             mock_graph.add_node.assert_not_called()
#             mock_graph.add_edge.assert_not_called()
#
#     @patch("lionagi.libs.SysUtil.check_import", side_effect=ImportError)
#     def test_to_networkx_import_error(self, mock_check_import):
#         with self.assertRaises(ImportError):
#             self.graph.to_networkx()
#         mock_check_import.assert_called_once_with("networkx")
#
#
# if __name__ == "__main__":
#     unittest.main()



"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import asyncio
import numpy as np
from dotenv import load_dotenv
from lionagi.libs import SysUtil, BaseService, StatusTracker, APIUtil, to_list, ninsert
from .abc import Component, ModelLimitExceededError

load_dotenv()


_oai_price_map = {
    "gpt-4o": (5, 15),
    "gpt-4-turbo": (10, 30),
    "gpt-3.5-turbo": (0.5, 1.5),
}


class iModel:
    """
    iModel is a class for managing AI model configurations and service
    integrations.

    Attributes:
        ln_id (str): A unique identifier for the model instance.
        timestamp (str): The timestamp when the model instance is created.
        endpoint (str): The API endpoint for the model service.
        provider_schema (dict): The schema for the service provider.
        provider (BaseService): The service provider instance.
        endpoint_schema (dict): The schema for the endpoint configuration.
        api_key (str): The API key for the service provider.
        status_tracker (StatusTracker): Instance of StatusTracker to track
            service status.
        service (BaseService): Configured service instance.
        config (dict): Configuration dictionary for the model.
        iModel_name (str): Name of the model.
    """

    default_model = "gpt-4o"

    def __init__(
        self,
        model: str = None,
        config: dict = None,
        provider: str = None,
        provider_schema: dict = None,
        endpoint: str = "chat/completions",
        token_encoding_name: str = None,
        api_key: str = None,
        api_key_schema: str = None,
        interval_tokens: int = None,
        interval_requests: int = None,
        interval: int = None,
        service: BaseService = None,
        allowed_parameters=[],
        device: str = None,
        costs=None,
        **kwargs,  # additional parameters for the model
    ):
        """
        Initializes an instance of the iModel class.

        Args:
            model (str, optional): Name of the model.
            config (dict, optional): Configuration dictionary.
            provider (str, optional): Name or class of the provider.
            provider_schema (dict, optional): Schema dictionary for the
                provider.
            endpoint (str, optional): Endpoint string, default is
                "chat/completions".
            token_encoding_name (str, optional): Name of the token encoding,
                default is "cl100k_base".
            api_key (str, optional): API key for the provider.
            api_key_schema (str, optional): Schema for the API key.
            interval_tokens (int, optional): Token interval limit, default is
                100,000.
            interval_requests (int, optional): Request interval limit, default
                is 1,000.
            interval (int, optional): Time interval in seconds, default is 60.
            service (BaseService, optional): An instance of BaseService.
            **kwargs: Additional parameters for the model.
        """
        self.ln_id: str = SysUtil.create_id()
        self.timestamp: str = SysUtil.get_timestamp(sep=None)[:-6]
        self.endpoint = endpoint
        self.allowed_parameters = allowed_parameters
        if isinstance(provider, type):
            provider = provider.__name__.replace("Service", "").lower()

        else:
            provider = str(provider).lower() if provider else "openai"

        from lionagi.integrations.provider._mapping import (
            SERVICE_PROVIDERS_MAPPING,
        )

        self.provider_schema = (
            provider_schema or SERVICE_PROVIDERS_MAPPING[provider]["schema"]
        )
        self.provider = SERVICE_PROVIDERS_MAPPING[provider]["service"]
        self.endpoint_schema = self.provider_schema.get(endpoint, {})
        self.token_limit = self.endpoint_schema.get("token_limit", 100_000)

        if api_key is not None:
            self.api_key = api_key

        elif api_key_schema is not None:
            self.api_key = os.getenv(api_key_schema)
        else:
            api_schema = self.provider_schema.get("API_key_schema", None)
            if api_schema:
                self.api_key = os.getenv(
                    self.provider_schema["API_key_schema"][0], None
                )

        self.status_tracker = StatusTracker()

        set_up_kwargs = {
            "api_key": getattr(self, "api_key", None),
            "schema": self.provider_schema or None,
            "endpoint": self.endpoint,
            "token_limit": self.token_limit,
            "token_encoding_name": token_encoding_name
            or self.endpoint_schema.get("token_encoding_name", None),
            "max_tokens": interval_tokens
            or self.endpoint_schema.get("interval_tokens", None),
            "max_requests": interval_requests
            or self.endpoint_schema.get("interval_requests", None),
            "interval": interval or self.endpoint_schema.get("interval", None),
        }

        set_up_kwargs = {
            k: v
            for k, v in set_up_kwargs.items()
            if v is not None and k in self.allowed_parameters
        }

        self.config = self._set_up_params(
            config or self.endpoint_schema.get("config", {}), **kwargs
        )

        if not model:
            if "model" not in self.config:
                model = SERVICE_PROVIDERS_MAPPING[provider]["default_model"]

        if self.config.get("model", None) != model and model is not None:
            self.iModel_name = model
            self.config["model"] = model
            ninsert(self.endpoint_schema, ["config", "model"], model)

        else:
            self.iModel_name = self.config["model"]

        if device:
            set_up_kwargs["device"] = device
        set_up_kwargs["model"] = self.iModel_name
        self.service: BaseService = self._set_up_service(
            service=service,
            provider=self.provider,
            **set_up_kwargs,
        )
        if self.iModel_name in _oai_price_map:
            self.costs = _oai_price_map[self.iModel_name]
        else:
            self.costs = costs or (0, 0)

    def update_config(self, **kwargs):
        """
        Updates the configuration with additional parameters.

        Args:
            **kwargs: Additional parameters to update the configuration.
        """
        self.config = self._set_up_params(self.config, **kwargs)

    def _set_up_config(self, model_config, **kwargs):
        """
        Sets up the model configuration.

        Args:
            model_config (dict): The default configuration dictionary.
            **kwargs: Additional parameters to update the configuration.

        Returns:
            dict: Updated configuration dictionary.
        """
        return {**model_config, **kwargs}

    def _set_up_service(self, service=None, provider=None, **kwargs):
        """
        Sets up the service for the model.

        Args:
            service (BaseService, optional): An instance of BaseService.
            provider (str, optional): Provider name or instance.
            **kwargs: Additional parameters for the service.

        Returns:
            BaseService: Configured service instance.
        """
        if not service:
            provider = provider or self.provider
            a = provider.__name__.replace("Service", "").lower()
            if a in ["openai", "openrouter"]:
                kwargs.pop("model", None)

            return provider(**kwargs)
        return service

    def _set_up_params(self, default_config=None, **kwargs):
        """
        Sets up the parameters for the model.

        Args:
            default_config (dict, optional): The default configuration
                dictionary.
            **kwargs: Additional parameters to update the configuration.

        Returns:
            dict: Updated parameters dictionary.

        Raises:
            ValueError: If any parameter is not allowed.
        """
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        params = {**default_config, **kwargs}
        allowed_params = self.endpoint_schema.get(
            "required", []
        ) + self.endpoint_schema.get("optional", [])

        if allowed_params != []:
            if (
                len(
                    not_allowed := [k for k in params.keys() if k not in allowed_params]
                )
                > 0
            ):
                raise ValueError(f"Not allowed parameters: {not_allowed}")

        return params

    async def call_chat_completion(self, messages, **kwargs):
        """
        Asynchronous method to call the chat completion service.

        Args:
            messages (list): List of messages for the chat completion.
            **kwargs: Additional parameters for the service call.

        Returns:
            dict: Response from the chat completion service.
        """

        num_tokens = APIUtil.calculate_num_token(
            {"messages": messages},
            "chat/completions",
            self.endpoint_schema.get("token_encoding_name", None),
        )

        if num_tokens > self.token_limit:
            raise ModelLimitExceededError(
                f"Number of tokens {num_tokens} exceeds the limit {self.token_limit}"
            )

        return await self.service.serve_chat(
            messages, required_tokens=num_tokens, **kwargs
        )

    async def call_embedding(self, embed_str, **kwargs):
        """
        Asynchronous method to call the embedding service.

        Args:
            input_file (str): Path to the input file.
            **kwargs: Additional parameters for the service call.

        Returns:
            dict: Response from the embedding service.
        """
        return await self.service.serve_embedding(embed_str, **kwargs)

    async def embed_node(self, node, field="content", **kwargs) -> bool:
        """
        if not specify field, we embed node.content
        """
        if not isinstance(node, Component):
            raise ValueError("Node must a lionagi item")
        embed_str = getattr(node, field)

        if isinstance(embed_str, dict) and "images" in embed_str:
            embed_str.pop("images", None)
            embed_str.pop("image_detail", None)

        num_tokens = APIUtil.calculate_num_token(
            {"input": str(embed_str) if isinstance(embed_str, dict) else embed_str},
            "embeddings",
            self.endpoint_schema["token_encoding_name"],
        )

        if self.token_limit and num_tokens > self.token_limit:
            raise ModelLimitExceededError(
                f"Number of tokens {num_tokens} exceeds the limit {self.token_limit}"
            )

        payload, embed = await self.call_embedding(embed_str, **kwargs)
        payload.pop("input")
        node.add_field("embedding", embed["data"][0]["embedding"])
        node._meta_insert("embedding_meta", payload)

    def to_dict(self):
        """
        Converts the model instance to a dictionary representation.

        Returns:
            dict: Dictionary representation of the model instance.
        """
        return {
            "ln_id": self.ln_id,
            "timestamp": self.timestamp,
            "provider": self.provider.__name__.replace("Service", ""),
            "api_key": self.api_key[:4]
            + "*" * (len(self.api_key) - 8)
            + self.api_key[-4:],
            "endpoint": self.endpoint,
            "token_encoding_name": self.service.token_encoding_name,
            **{
                k: v
                for k, v in self.config.items()
                if k in getattr(self.service, "allowed_kwargs", []) and v is not None
            },
            "model_costs": None if self.costs == (0, 0) else self.costs,
        }

    async def compute_perplexity(
        self,
        initial_context: str = None,
        tokens: list[str] = None,
        system_msg: str = None,
        n_samples: int = 1,  # number of samples used for the computation
        use_residue: bool = True,  # whether to use residue for the last sample
        **kwargs,  # additional arguments for the model
    ) -> tuple[list[str], float]:
        tasks = []
        context = initial_context or ""

        n_samples = n_samples or len(tokens)
        sample_token_len, residue = divmod(len(tokens), n_samples)
        samples = []

        if n_samples == 1:
            samples = [tokens]
        else:
            samples = [tokens[: (i + 1) * sample_token_len] for i in range(n_samples)]

            if use_residue and residue != 0:
                samples.append(tokens[-residue:])

        sampless = [context + " ".join(sample) for sample in samples]

        for sample in sampless:
            messages = [{"role": "system", "content": system_msg}] if system_msg else []
            messages.append(
                {"role": "user", "content": sample},
            )

            task = asyncio.create_task(
                self.call_chat_completion(
                    messages=messages,
                    logprobs=True,
                    max_tokens=sample_token_len,
                    **kwargs,
                )
            )
            tasks.append(task)

        results = await asyncio.gather(*tasks)  # result is (payload, response)
        results_ = []
        num_prompt_tokens = 0
        num_completion_tokens = 0

        for idx, result in enumerate(results):
            _dict = {}
            _dict["tokens"] = samples[idx]

            num_prompt_tokens += result[1]["usage"]["prompt_tokens"]
            num_completion_tokens += result[1]["usage"]["completion_tokens"]

            logprobs = result[1]["choices"][0]["logprobs"]["content"]
            logprobs = to_list(logprobs, flatten=True, dropna=True)
            _dict["logprobs"] = [(i["token"], i["logprob"]) for i in logprobs]
            results_.append(_dict)

        logprobs = to_list(
            [[i[1] for i in d["logprobs"]] for d in results_], flatten=True
        )

        return {
            "tokens": tokens,
            "n_samples": n_samples,
            "num_prompt_tokens": num_prompt_tokens,
            "num_completion_tokens": num_completion_tokens,
            "logprobs": logprobs,
            "perplexity": np.exp(np.mean(logprobs)),
        }



"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Any
from pydantic import Field
from .message import RoledMessage, MessageRole
from .action_request import ActionRequest


# action response must correlates to a specific action request
class ActionResponse(RoledMessage):
    """
    Represents a response to a specific action request.

    Inherits from `RoledMessage` and provides attributes specific to action responses.

    Attributes:
        action_request (str): The ID of the action request that this response corresponds to.
        function (str): The name of the function called.
        arguments (dict): The keyword arguments provided.
        func_outputs (Any): The output of the function call.
    """

    action_request: str | None = Field(
        None,
        description="The id of the action request that this response corresponds to",
    )

    function: str | None = Field(None, description="The name of the function called")
    arguments: dict | None = Field(None, description="The keyword arguments provided")
    func_outputs: Any | None = Field(
        None, description="The output of the function call"
    )

    def __init__(
        self,
        action_request: ActionRequest,
        sender: str | None = None,  # the sender of action request
        func_outputs=None,
        **kwargs,
    ):
        """
        Initializes the ActionResponse.

        Args:
            action_request (ActionRequest): The action request that this response corresponds to.
            sender (str, optional): The sender of the action request.
            func_outputs (Any, optional): The output of the function call.

        Raises:
            ValueError: If the action request has already been responded to.
        """
        if action_request.is_responded():
            raise ValueError("Action request has already been responded to")

        super().__init__(
            role=MessageRole.ASSISTANT,
            sender=sender or "N/A",  # sender is the actionable component
            recipient=action_request.sender,  # recipient is the assistant who made the request
            content={
                "action_response": {
                    "function": action_request.function,
                    "arguments": action_request.arguments,
                    "output": func_outputs,
                }
            },
            **kwargs,
        )
        self.update_request(action_request)
        self.func_outputs = func_outputs

    def update_request(self, action_request: ActionRequest):
        """
        Updates the action request details in the action response.

        Args:
            action_request (ActionRequest): The action request to update from.
        """
        self.function = action_request.function
        self.arguments = action_request.arguments
        self.action_request = action_request.ln_id
        action_request.action_response = self.ln_id

    def _to_dict(self):
        """
        Converts the action response to a dictionary.

        Returns:
            dict: A dictionary representation of the action response.
        """
        return {
            "function": self.function,
            "arguments": self.arguments,
            "output": self.func_outputs,
        }

    def clone(self, **kwargs):
        """
        Creates a copy of the current object with optional additional arguments.

        This method clones the current object, preserving its function and arguments.
        It also retains the original `action_request`, `func_outputs`, and metadata,
        while allowing for the addition of new attributes through keyword arguments.

        Args:
            **kwargs: Optional keyword arguments to be included in the cloned object.

        Returns:
            ActionResponse: A new instance of the object with the same function, arguments,
            and additional keyword arguments.
        """
        import json

        arguments = json.dumps(self.arguments)
        action_request = ActionRequest(
            function=self.function, arguments=json.loads(arguments)
        )
        action_response_copy = ActionResponse(action_request=action_request, **kwargs)
        action_response_copy.action_request = self.action_request
        action_response_copy.func_outputs = self.func_outputs
        action_response_copy.metadata["origin_ln_id"] = self.ln_id
        return action_response_copy



import unittest
from unittest.mock import patch
from lionagi.libs.ln_validate import (
    check_dict_field,
    check_action_field,
    check_number_field,
    check_bool_field,
    check_str_field,
    check_enum_field,
    _fix_action_field,
    _fix_number_field,
    _fix_bool_field,
    _fix_str_field,
    _fix_enum_field,
)


class TestValidationFunctions(unittest.TestCase):
    def test_check_dict_field_valid(self):
        x = {"key": "value"}
        keys = ["key"]
        result = check_dict_field(x, keys)
        self.assertEqual(result, x)

    def test_fix_action_field_invalid_string(self):
        x = '[{"invalid_key": "value"}]'
        try:
            _fix_action_field(x)
        except ValueError as e:
            self.assertEqual(str(e), "Invalid action field: {'invalid_key': 'value'}")

    def test_fix_action_field_invalid_discard_disabled(self):
        x = '[{"function": "func1", "arguments": {}}, {"invalid_key": "value"}]'
        try:
            _fix_action_field(x, discard_=False)
        except ValueError as e:
            self.assertEqual(str(e), "Invalid action field: {'invalid_key': 'value'}")

    def test_check_dict_field_invalid_fixable(self):
        x = '{"key": "value"}'
        keys = ["key"]
        result = check_dict_field(x, keys, fix_=True)
        self.assertEqual(result, {"key": "value"})

    def test_check_dict_field_invalid_not_fixable(self):
        x = '{"invalid_key": "value"}'
        keys = ["key"]
        with self.assertRaises(ValueError) as cm:
            check_dict_field(x, keys, fix_=False)
        self.assertEqual(
            str(cm.exception), "Default value for DICT must be a dict, got str"
        )

    def test_check_action_field_invalid_not_fixable(self):
        x = '[{"invalid_key": "value"}]'
        try:
            check_action_field(x, fix_=False)
        except ValueError as e:
            self.assertEqual(str(e), "Invalid action field type.")

    def test_check_action_field_invalid_fix_disabled(self):
        x = '[{"function": "func1", "arguments": {}}, {"invalid_key": "value"}]'
        try:
            check_action_field(x, fix_=False)
        except ValueError as e:
            self.assertEqual(str(e), "Invalid action field type.")

    def test_check_action_field_valid(self):
        x = [
            {"function": "func1", "arguments": {}},
            {"function": "func2", "arguments": {}},
        ]
        result = check_action_field(x)
        self.assertEqual(result, x)

    def test_check_action_field_invalid_fixable(self):
        x = '[{"function": "func1", "arguments": {}}, {"function": "func2", "arguments": {}}]'
        result = check_action_field(x, fix_=True)
        self.assertEqual(
            result,
            [
                {"function": "func1", "arguments": {}},
                {"function": "func2", "arguments": {}},
            ],
        )

    def test_check_action_field_invalid_fix_disabled(self):
        x = '[{"function": "func1", "arguments": {}}, {"invalid_key": "value"}]'
        try:
            check_action_field(x, fix_=False)
        except ValueError as e:
            self.assertEqual(str(e), "Invalid action field type.")

    def test_check_number_field_valid(self):
        x = 42
        result = check_number_field(x)
        self.assertEqual(result, x)

    def test_check_number_field_invalid_fixable(self):
        x = "42"
        result = check_number_field(x, fix_=True)
        self.assertEqual(result, 42)

    def test_check_number_field_invalid_not_fixable(self):
        x = "not_a_number"
        with self.assertRaises(ValueError) as cm:
            check_number_field(x, fix_=True)
        self.assertEqual(
            str(cm.exception), "Failed to convert not_a_number into a numeric value"
        )

    def test_check_number_field_invalid_fix_disabled(self):
        x = "42"
        with self.assertRaises(ValueError) as cm:
            check_number_field(x, fix_=False)
        self.assertEqual(
            str(cm.exception),
            "Default value for NUMERIC must be an int or float, got str",
        )

    def test_check_bool_field_valid(self):
        x = True
        result = check_bool_field(x)
        self.assertEqual(result, x)

    def test_check_bool_field_invalid_fixable(self):
        x = "true"
        result = check_bool_field(x, fix_=True)
        self.assertTrue(result)

    def test_check_bool_field_invalid_not_fixable(self):
        x = "not_a_boolean"
        with self.assertRaises(ValueError) as cm:
            check_bool_field(x, fix_=True)
        self.assertEqual(
            str(cm.exception), "Failed to convert not_a_boolean into a boolean value"
        )

    def test_check_bool_field_invalid_fix_disabled(self):
        x = "true"
        with self.assertRaises(ValueError) as cm:
            check_bool_field(x, fix_=False)
        self.assertEqual(
            str(cm.exception), "Default value for BOOLEAN must be a bool, got str"
        )

    def test_check_str_field_valid(self):
        x = "hello"
        result = check_str_field(x)
        self.assertEqual(result, x)

    def test_check_str_field_invalid_fixable(self):
        x = 42
        result = check_str_field(x, fix_=True)
        self.assertEqual(result, "42")

    def test_check_str_field_invalid_not_fixable(self):
        x = object()
        try:
            check_str_field(x, fix_=False)
        except ValueError as e:
            self.assertEqual(
                str(e), "Default value for STRING must be a str, got object"
            )

    def test_check_str_field_invalid_fix_disabled(self):
        x = 42
        with self.assertRaises(ValueError) as cm:
            check_str_field(x, fix_=False)
        self.assertEqual(
            str(cm.exception), "Default value for STRING must be a str, got int"
        )

    def test_check_enum_field_valid(self):
        x = "option1"
        choices = ["option1", "option2"]
        result = check_enum_field(x, choices)
        self.assertEqual(result, x)

    def test_check_enum_field_invalid_fixable(self):
        x = "option3"
        choices = ["option1", "option2"]
        with patch("lionagi.libs.ln_validate._fix_enum_field", return_value="option1"):
            result = check_enum_field(x, choices, fix_=True)
        self.assertEqual(result, "option1")

    def test_check_enum_field_invalid_not_fixable(self):
        x = "option3"
        choices = ["option1", "option2"]
        with self.assertRaises(ValueError) as cm:
            check_enum_field(x, choices, fix_=False)
        self.assertEqual(
            str(cm.exception),
            "Default value for ENUM must be one of the ['option1', 'option2'], got option3",
        )

    def test_check_enum_field_invalid_choices(self):
        x = "option1"
        choices = ["option1", 42]
        with self.assertRaises(ValueError) as cm:
            check_enum_field(x, choices)
        self.assertEqual(
            str(cm.exception),
            "Field type ENUM requires all choices to be of the same type, got ['option1', 42]",
        )

    def test_check_enum_field_invalid_type(self):
        x = 42
        choices = ["option1", "option2"]
        with self.assertRaises(ValueError) as cm:
            check_enum_field(x, choices)
        self.assertEqual(
            str(cm.exception),
            "Default value for ENUM must be an instance of the str, got int",
        )

    def test_fix_action_field_string(self):
        x = '[{"function": "func1", "arguments": {}}, {"function": "func2", "arguments": {}}]'
        result = _fix_action_field(x)
        self.assertEqual(
            result,
            [
                {"function": "func1", "arguments": {}},
                {"function": "func2", "arguments": {}},
            ],
        )

    def test_fix_action_field_invalid_string(self):
        x = '[{"invalid_key": "value"}]'
        try:
            _fix_action_field(x)
        except ValueError as e:
            self.assertEqual(str(e), "Invalid action field: {'invalid_key': 'value'}")

    def test_fix_action_field_invalid_discard_disabled(self):
        x = '[{"function": "func1", "arguments": {}}, {"invalid_key": "value"}]'
        try:
            _fix_action_field(x, discard_=False)
        except ValueError as e:
            self.assertEqual(str(e), "Invalid action field: {'invalid_key': 'value'}")

    def test_fix_number_field_valid(self):
        x = "42"
        result = _fix_number_field(x)
        self.assertEqual(result, 42)

    def test_fix_number_field_invalid(self):
        x = "not_a_number"
        with self.assertRaises(ValueError) as cm:
            _fix_number_field(x)
        self.assertEqual(
            str(cm.exception), "Failed to convert not_a_number into a numeric value"
        )

    def test_fix_bool_field_true(self):
        x = "true"
        result = _fix_bool_field(x)
        self.assertTrue(result)

    def test_fix_bool_field_false(self):
        x = "false"
        result = _fix_bool_field(x)
        self.assertFalse(result)

    def test_fix_bool_field_invalid(self):
        x = "not_a_boolean"
        with self.assertRaises(ValueError) as cm:
            _fix_bool_field(x)
        self.assertEqual(
            str(cm.exception), "Failed to convert not_a_boolean into a boolean value"
        )

    def test_fix_str_field_valid(self):
        x = 42
        result = _fix_str_field(x)
        self.assertEqual(result, "42")

    def test_fix_str_field_invalid(self):
        x = object()
        try:
            _fix_str_field(x)
        except ValueError as e:
            self.assertEqual(
                str(e),
                "Failed to convert <object object at 0x{:x}> into a string value".format(
                    id(x)
                ),
            )

    def test_fix_enum_field_valid(self):
        x = "option3"
        choices = ["option1", "option2"]
        with patch(
            "lionagi.libs.StringMatch.choose_most_similar", return_value="option1"
        ):
            result = _fix_enum_field(x, choices)
        self.assertEqual(result, "option1")

    def test_fix_enum_field_invalid(self):
        x = "option3"
        choices = ["option1", "option2"]
        with patch(
            "lionagi.libs.StringMatch.choose_most_similar",
            side_effect=ValueError("No match found"),
        ):
            with self.assertRaises(ValueError) as cm:
                _fix_enum_field(x, choices)
        self.assertEqual(
            str(cm.exception), "Failed to convert option3 into one of the choices"
        )

    def test_fix_bool_field_false(self):
        x = "false"
        result = _fix_bool_field(x)
        self.assertFalse(result)

    def test_fix_str_field_invalid(self):
        x = object()
        try:
            _fix_str_field(x)
        except ValueError as e:
            self.assertEqual(
                str(e),
                "Failed to convert <object object at 0x{:x}> into a string value".format(
                    id(x)
                ),
            )

    def test_fix_enum_field_valid(self):
        x = "option3"
        choices = ["option1", "option2"]
        with patch(
            "lionagi.libs.StringMatch.choose_most_similar", return_value="option1"
        ):
            result = _fix_enum_field(x, choices)
        self.assertEqual(result, "option1")

    def test_fix_enum_field_invalid(self):
        x = "option3"
        choices = ["option1", "option2"]
        with patch(
            "lionagi.libs.StringMatch.choose_most_similar",
            side_effect=ValueError("No match found"),
        ):
            with self.assertRaises(ValueError) as cm:
                _fix_enum_field(x, choices)
        self.assertEqual(
            str(cm.exception), "Failed to convert option3 into one of the choices"
        )


if __name__ == "__main__":
    unittest.main()



"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import asyncio
import logging

from lionagi.libs.ln_func_call import rcall
from lionagi.core.work.worklog import WorkLog


class WorkFunction:
    """
    A class representing a work function.

    Attributes:
        assignment (str): The assignment description of the work function.
        function (Callable): The function to be performed.
        retry_kwargs (dict): The retry arguments for the function.
        worklog (WorkLog): The work log for the function.
        guidance (str): The guidance or documentation for the function.
    """

    def __init__(
        self, assignment, function, retry_kwargs=None, guidance=None, capacity=10, refresh_time=1
    ):
        """
        Initializes a WorkFunction instance.

        Args:
            assignment (str): The assignment description of the work function.
            function (Callable): The function to be performed.
            retry_kwargs (dict, optional): The retry arguments for the function.
                Defaults to None.
            guidance (str, optional): The guidance or documentation for the function.
                Defaults to None.
            capacity (int, optional): The capacity of the work queue batch processing.
                Defaults to 10.
            refresh_time (int, optional): The time interval to refresh the work log queue.
                Defaults to 1.
        """
        self.assignment = assignment
        self.function = function
        self.retry_kwargs = retry_kwargs or {}
        self.worklog = WorkLog(capacity, refresh_time=refresh_time)
        self.guidance = guidance or self.function.__doc__

    @property
    def name(self):
        """
        Gets the name of the function.

        Returns:
            str: The name of the function.
        """
        return self.function.__name__

    @property
    def execution_mode(self):
        """
         Gets the execution mode of the work function's queue.

         Returns:
             bool: The execution mode of the work function's queue.
         """
        return self.worklog.queue.execution_mode

    def is_progressable(self):
        """
        Checks if the work function is progressable.

        Returns:
            bool: True if the work function is progressable, otherwise False.
        """
        return self.worklog.pending_work and not self.worklog.stopped

    async def perform(self, *args, **kwargs):
        """
        Performs the work function with retry logic.

        Args:
            *args: Positional arguments for the function.
            **kwargs: Keyword arguments for the function.

        Returns:
            Any: The result of the function call.
        """
        kwargs = {**self.retry_kwargs, **kwargs}
        return await rcall(self.function, *args, timing=True, **kwargs)

    async def forward(self):
        """
        Forward the work log to work queue.
        """
        await self.worklog.forward()

    async def process(self):
        """
        Process the first capacity_size works in the work queue.
        """
        await self.worklog.queue.process()

    async def execute(self):
        """
        Starts the execution of the work function's queue.
        """
        asyncio.create_task(self.worklog.queue.execute())

    async def stop(self):
        """
        Stops the execution of the work function's queue.
        """
        await self.worklog.stop()


"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# import logging

# # Configure logging
# logging.basicConfig(
#     level=logging.ERROR,
#     format="%(asctime)s:%(levelname)s:%(message)s",
# )


class LionAGIError(Exception):
    """Base class for all exceptions in the LionAGI system."""

    def __init__(self, message=None):
        if message is None:
            message = "An unspecified error occurred in the LionAGI system."
        super().__init__(message)


class LionValueError(LionAGIError):
    """Exception raised for errors in the input value."""

    def __init__(self, message=None):
        if message is None:
            message = "An error occurred in the input value."
        super().__init__(message)


class LionTypeError(LionAGIError):
    """Exception raised for type mismatch or type checking errors."""

    def __init__(self, message=None):
        if message is None:
            message = "Item must be identifiable, `ln_id` or `Component`"
        super().__init__(message)


class LionItemError(LionAGIError):
    """Base class for exceptions related to LionAGI items."""

    def __init__(self, item, message=None):
        if message is None:
            message = "An error occurred with the specified item."
        super().__init__(f"{message} Item: '{item}'.")


class ItemNotFoundError(LionItemError):
    """Exception raised when a specified item is not found."""

    def __init__(self, item):
        super().__init__(item, "Item not found.")


class ItemInvalidError(LionItemError):
    """Exception raised when an invalid item is used in an operation."""

    def __init__(self, item):
        super().__init__(item, "The item is invalid for this operation.")


class FieldError(LionAGIError):
    """Exception raised for errors in field validation."""

    def __init__(self, field, message=None):
        if message is None:
            message = "An error occurred with the specified field."
        super().__init__(f"{message}: {field}.")


class LionOperationError(LionAGIError):
    """Base class for exceptions related to operational failures."""

    def __init__(self, operation, message=None):
        if message is None:
            message = "An operation failed."
        super().__init__(f"{message} Operation: '{operation}'.")


class ConcurrencyError(LionOperationError):
    """Exception raised for errors due to concurrency issues."""

    def __init__(self, operation=None):
        if operation is None:
            operation = "a concurrent operation"
        super().__init__(operation, "A concurrency error occurred during")


class RelationError(LionAGIError):
    """Exception raised for errors in relation operations."""

    def __init__(self, message=None):
        if message is None:
            message = "Nodes are not related."
        super().__init__(message)


class ActionError(LionAGIError):
    """Exception raised for errors in action operations."""

    def __init__(self, message=None):
        if message is None:
            message = "An error occurred with the specified action."
        super().__init__(message)


class ModelLimitExceededError(LionOperationError):
    """Exception raised when a resource limit is exceeded."""

    def __init__(self, message=None):
        if message is None:
            message = "The model limit has been exceeded."
        super().__init__("Model", message)


class TimeoutError(LionOperationError):
    """Exception raised when an operation times out."""

    def __init__(self, operation, timeout):
        super().__init__(operation, f"Operation timed out after {timeout} seconds.")


class ServiceError(LionAGIError):
    """Exception raised for errors in endpoint configuration."""

    def __init__(self, message, errors=None):
        super().__init__(message, errors)

    @classmethod
    def unavailable(cls, endpoint, service, err_msg=None):
        msg = f"{endpoint} is currently unavailable"
        if err_msg:
            msg += f": {err_msg}"
        else:
            msg += f" for {service.__class__.__name__}"
        return cls(msg)
